from shutil import which

from coalib.bearlib.abstractions.Linter import linter
from coalib.bearlib import deprecate_settings
from coala_utils.param_conversion import negate


@linter('bash', output_format='regex',
        output_regex=r'.+:(?P<line>.+):(?P<message>.*)')
class JavaPMDBear:
    """
    Check Java code for possible issues like potential bugs, dead code or too
    complicated expressions.

    More information is available at
    <http://pmd.github.io/pmd-5.4.1/pmd-java/rules/index.html>.
    """

    LANGUAGES = {'Java'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Code Simplification', 'Unreachable Code', 'Smell',
                  'Duplication'}

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which('bash') is None:
            return 'bash is not installed.'
        elif which('pmd') is None and which('run.sh') is None:
            return ('PMD is missing. Make sure to install it from '
                    '<https://pmd.github.io/>')
        else:
            return True

    @staticmethod
    @deprecate_settings(allow_unnecessary_code=('check_unnecessary', negate),
                        allow_unused_code=('check_unused', negate))
    def create_arguments(filename, file, config_file,
                         check_best_practices: bool = True,
                         check_braces: bool = True,
                         check_clone_implementation: bool = True,
                         check_code_size: bool = True,
                         check_comments: bool = False,
                         check_controversial: bool = False,
                         check_design: bool = False,
                         check_imports: bool = True, check_naming: bool = True,
                         check_optimizations: bool = False,
                         check_strings: bool = False,
                         allow_unnecessary_code: bool = False,
                         allow_unused_code: bool = False):
        """
        :param check_best_practices:
            Checks for best practices.
        :param check_braces:
            Checks for the right use of braces.
        :param check_clone_implementation:
            Checks for the right implementation of the ``clone()`` function.
        :param check_code_size:
            Checks for large or complicated code constructs.
        :param check_comments:
            Checks comments for length, content and placement.
        :param check_controversial:
            Does various checks that are considered controversial.
        :param check_design:
            Checks for optimal code implementations. Alternate approaches
            are suggested.
        :param check_imports:
            Checks for duplicate and unused imports.
        :param check_naming:
            Checks the names of identifiers against some rules.
        :param check_optimizations:
            Checks for best pratices regarding optimization.
        :param check_strings:
            Checks for String, StringBuffer and StringBuilder instances.
        :param allow_unnecessary_code:
            Allows unnecessary code.
        :param allow_unused_code:
            Allows unused code.
        """
        options = {
            'java-basic': check_best_practices,
            'java-braces': check_braces,
            'java-clone': check_clone_implementation,
            'java-codesize': check_code_size,
            'java-comments': check_comments,
            'java-controversial': check_controversial,
            'java-design': check_design,
            'java-imports': check_imports,
            'java-naming': check_naming,
            'java-optimizations': check_optimizations,
            'java-strings': check_strings,
            'java-unnecessary': not allow_unnecessary_code,
            'java-unusedcode': not allow_unused_code}
        rules = ','.join(key for key in options if options[key])

        executable = which('pmd') or which('run.sh')  # Mac vs. Unix
        return executable, 'pmd', '-R', rules, '-d', filename



import itertools
import re

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.settings.Setting import path, url
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


def path_or_url(xml_dtd):
    """
    Converts the setting value to url or path.

    :param xml_dtd: Setting key.
    :return:        Returns a converted setting value.
    """
    try:
        return url(xml_dtd)
    except ValueError:
        return path(xml_dtd)


@linter(executable='xmllint',
        use_stdout=True,
        use_stderr=True)
class XMLBear:
    """
    Checks the code with ``xmllint``.

    See http://xmlsoft.org/xmllint.html
    """
    LANGUAGES = {'XML'}
    REQUIREMENTS = {
        DistributionRequirement(
            'libxml2',
            apt_get='libxml2-utils',
            portage='dev-libs/libxml2',
        ),
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}

    _output_regex = re.compile(
        r'.*:(?P<line>\d+):.*(?P<severity>error|warning)\s?: '
        r'(?P<message>.*)\n.*\n.*')

    @staticmethod
    def create_arguments(filename, file, config_file,
                         xml_schema: path='',
                         xml_dtd: path_or_url=''):
        """
        :param xml_schema: ``W3C XML Schema`` file used for validation.
        :param xml_dtd:    ``Document type Definition (DTD)`` file or
                           url used for validation.
        """
        args = (filename,)
        if xml_schema:
            args += ('-schema', xml_schema)
        if xml_dtd:
            args += ('-dtdvalid', xml_dtd)

        return args

    def process_output(self, output, filename, file):
        stdout, stderr = output
        if stdout:
            # Return issues from stderr and stdout if stdout is not empty
            return itertools.chain(
                self.process_output_regex(
                    stderr, filename, file,
                    output_regex=self._output_regex),
                self.process_output_corrected(
                    stdout, filename, file,
                    diff_severity=RESULT_SEVERITY.INFO,
                    result_message='XML can be formatted better.'))
        else:
            # Return issues from stderr if stdout is empty
            return self.process_output_regex(
                stderr, filename, file,
                output_regex=self._output_regex)



import os.path

from coalib.bears.LocalBear import LocalBear
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.bearlib.naming_conventions import (
    to_camelcase, to_kebabcase, to_pascalcase, to_snakecase, to_spacecase)


class FilenameBear(LocalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'

    _naming_convention = {'camel': to_camelcase,
                          'kebab': to_kebabcase,
                          'pascal': to_pascalcase,
                          'snake': to_snakecase,
                          'space': to_spacecase}

    _language_naming_convention = {
                                   '.java': 'pascal',
                                   '.js': 'kebab',
                                   '.py': 'snake',
                                  }

    def run(self, filename, file,
            file_naming_convention: str=None,
            ignore_uppercase_filenames: bool=True):
        """
        Checks whether the filename follows a certain naming-convention.

        :param file_naming_convention:
            The naming-convention. Supported values are:
            - ``auto`` to guess the correct convention. Defaults to ``snake``
            if the correct convention cannot be guessed.
            - ``camel`` (``thisIsCamelCase``)
            - ``kebab`` (``this-is-kebab-case``)
            - ``pascal`` (``ThisIsPascalCase``)
            - ``snake`` (``this_is_snake_case``)
            - ``space`` (``This Is Space Case``)
        :param ignore_uppercase_filenames:
            Whether or not to ignore fully uppercase filenames completely,
            e.g. COPYING, LICENSE etc.
        """
        head, tail = os.path.split(filename)
        filename_without_extension, extension = os.path.splitext(tail)

        if file_naming_convention is None:
            self.warn('Please specify a file naming convention explicitly'
                      ' or use "auto".')
            file_naming_convention = 'auto'
        else:
            file_naming_convention = file_naming_convention.lower()

        if file_naming_convention == 'auto':
            if extension in self._language_naming_convention:
                file_naming_convention = self._language_naming_convention[
                    extension]
            else:
                self.warn('The file naming convention could not be guessed. '
                          'Using the default "snake" naming convention.')
                file_naming_convention = 'snake'

        try:
            new_name = self._naming_convention[file_naming_convention](
                filename_without_extension)
        except KeyError:
            self.err('Invalid file-naming-convention provided: ' +
                     file_naming_convention)
            return

        if ignore_uppercase_filenames and filename_without_extension.isupper():
            return

        if new_name != filename_without_extension:
            diff = Diff(file, rename=os.path.join(head, new_name + extension))

            yield Result(
                self,
                'Filename does not follow {} naming-convention.'.format(
                    file_naming_convention),
                diff.affected_code(filename),
                diffs={filename: diff})



import yaml

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='scss-lint', output_format='regex',
        output_regex=r'.+:(?P<line>\d+)\s+\[(?P<severity>.)\]\s*'
                     r'(?P<message>.*)')
class SCSSLintBear:
    """
    Check SCSS code to keep it clean and readable.

    More information is available at <https://github.com/brigade/scss-lint>.
    """

    LANGUAGES = {'SCSS'}
    # require flag is necessary for 'scss_lint'
    # https://github.com/brigade/scss-lint#installation
    REQUIREMENTS = {GemRequirement('scss_lint', '', 'false'),
                    PipRequirement('pyyaml', '3.12')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename, '--config', config_file

    @staticmethod
    def generate_config(filename, file,
                        space_around_bang: list=[True, False],
                        allow_chained_classes: bool=False,
                        prefer_color_keywords: bool=False,
                        use_color_variables: bool=True,
                        allow_debug_statement: bool=False,
                        check_declaration_order: bool=True,
                        allow_duplicate_properties: bool=False,
                        allow_consecutives_duplicate_property: bool=False,
                        else_on_same_line: bool=True,
                        force_empty_line_between_blocks: bool=True,
                        allow_empty_rules: bool=False,
                        use_short_hexadecimal_length_style: bool=True,
                        use_lowercase_hexadecimal: bool=True,
                        validate_hexadecimal: bool=True,
                        allow_id_selector: bool=False,
                        allow_important_rule_in_properties: bool=False,
                        use_spaces: bool=True,
                        indent_size: int=2,
                        exclude_leading_zero: bool=True,
                        allow_mergeable_selectors: bool=False,
                        allow_leading_underscore: bool=True,
                        function_naming_convention: str='hyphen',
                        mixin_naming_convention: str='hyphen',
                        variable_naming_convention: str='hyphen',
                        placeholder_naming_convention: str='hyphen',
                        max_nesting_depth: int=3,
                        use_placeholder_selector_in_extend: bool=True,
                        max_properties: int=10,
                        allow_unit_on_zero_values: bool=False,
                        check_ulrs_format: bool=True,
                        urls_in_quotes: bool=True,
                        allow_unnecesseary_parent_reference: bool=False,
                        allow_unnecessary_mantissa: bool=False,
                        allow_trailing_whitespaces: bool=False,
                        allow_trailing_semicolon: bool=True,
                        check_imports_path: bool=True,
                        allow_filename_leading_underscore: bool=False,
                        allow_filename_extension: bool=False,
                        use_length_variables: bool=True,
                        check_properties_spelling: bool=True,
                        extra_properties: list=(),
                        disabled_properties: list=(),
                        check_pseudo_elements: bool=True,
                        spaces_between_parentheses: int=0,
                        spaces_around_operators: str=1):
        """
        :param space_around_bang:
            Enforces a space before and/or after ``!`` (the "bang").
        :param allow_chained_classes:
            Allows defining a rule set using a selector with chained classes.
        :param prefer_color_keywords:
            Prefers color keywords over hexadecimal color codes.
        :param use_color_variables:
            Prefers color literals (keywords or hexadecimal codes) to be used
            only in variable declarations.
        :param allow_debug_statement:
            Allows ``@debug`` statements.
        :param check_declaration_order:
            Rule sets should be ordered as follows: ``@extend`` declarations,
            ``@include`` declarations without inner ``@content``, properties,
            ``@include`` declarations with inner ``@content``, then nested rule
            sets.
        :param allow_duplicate_properties:
            Allows defining the same property twice in a single rule set.
        :param allow_consecutives_duplicate_property:
            Allows defining the same property consecutively in a single rule
            set.
        :param else_on_same_line:
            Places ``@else`` statements on the same line as the preceding curly
            brace.
        :param force_empty_line_between_blocks:
            Separate rule, function, and mixin declarations with empty lines.
        :param allow_empty_rules:
            Allows empty rule set.
        :param use_short_hexadecimal_length_style:
            Prefer shorthand or long-form hexadecimal colors by setting the
            style option to short or long, respectively.
        :param use_lowercase_hexadecimal:
            Checks if hexadecimal colors are written in lowercase or uppercase.
        :param validate_hexadecimal:
            Ensure hexadecimal colors are valid (either three or six digits).
        :param allow_id_selector:
            Allows using ID selectors.
        :param allow_important_rule_in_property:
            Allows using ``!important`` in properties.
        :param use_spaces:
            Use spaces for indentation (tabs otherwise).
        :param indent_size:
            Number of spaces per indentation level.
        :param exclude_leading_zero:
            Determines whether leading zeros should be written or not in
            numeric values with a decimal point.
        :param allow_mergeable_selectors:
            Allows defining the same selector twice in a single sheet.
        :param allow_leading_underscore:
            Allows names to start with a single underscore.
        :param function_naming_convention:
            Name of convention (``hyphen``(use lowercase letters and hyphens)
            (default), ``camel``, ``snake``), or a ``regex`` the name must
            match (eg: ``^[a-zA-Z]+$``) to use for functions.
        :param mixin_naming_convention:
            Name of convention (``hyphen`` (default), ``camel``, ``snake``), or
            a regex the name must match (eg: ``^[a-zA-Z]+$``) to use for
            mixins.
        :param variable_naming_convention:
            Name of convention (``hyphen`` (default), ``camel``, ``snake``), or
            a regex the name must match (eg: ``^[a-zA-Z]+$``) to use for
            variables.
        :param placeholder_naming_convention:
            Name of convention (``hyphen`` (default), ``camel``, ``snake``), or
            a regex the name must match (eg: ``^[a-zA-Z]+$``) to use for
            placeholders.
        :param max_nesting_depth:
            Maximum nesting depth.
        :param use_placeholder_selector_in_extend:
            Enforces using placeholder selectors in ``@extend``.
        :param max_properties:
            Enforces a limit on the number of properties in a rule set.
        :param allow_unit_on_zero_values:
            Allow omitting length units on zero values.
        :param check_urls_format:
            URLs should be valid and not contain protocols or domain names.
        :param urls_in_quotes:
            URLs should always be enclosed within quotes.
        :param allow_unnecessary_parent_reference:
            Allows use of the parent selector references ``&`` even when they
            are not unnecessary.
        :param allow_unnecessary_mantissa:
            Numeric values can contain unnecessary fractional portions.
        :param allow_traling_whitespaces:
            Unables trailing whitespace.
        :param allow_trailing_semicolon:
            Property values; ``@extend``, ``@include``, and ``@import``
            directives; and variable declarations should always end with a
            semicolon.
        :param check_imports_path:
            The basenames of ``@import``ed SCSS partials should not begin with
            an underscore and should not include the filename extension.
            These requirements can be modified by changing
            ``allow_filename_leading_underscore``, and ``allow_extensions``.
        :param allow_filename_leading_underscore:
            Requires basenames of ``@import``ed SCSS partials to begin with an
            underscore.  This setting require ``check_import_paths`` to be
            enabled.
        :param allow_filename_extension:
            Requires basenames of ``@import``ed SCSS partials to include
            filename extension, this setting require ``check_import_paths`` to
            be enabled.
        :param use_length_variables:
            Prefer length literals (numbers with units) to be used only in
            variable declarations.

            ::
                 div {
                   width: 100px;
                 }

            Is not valid, whereas

            ::
                 $column-width: 100px;

                 div {
                   width: $column-width;
                 }
            is valid.
        :param check_properties_spelling:
            Reports when an unknown or disabled CSS property is used
            (ignoring vendor-prefixed properties).
        :param extra_properties:
            List of extra properties to allow.
        :param disabled_properties:
            List of existing properties to deny.
        :param check_pseudo_elements:
            Pseudo-elements, like ``::before``, and ``::first-letter``,
            should be declared with two colons. Pseudo-classes, like ``:hover``
            and ``:first-child``, should be declared with one colon.

            ::
                p::before {
                  content: '>'
                }

                p:hover {
                  color: red;
                }

        :param spaces_between_parentheses:
            Spaces to require between parentheses.
        :param spaces_around_operators:
            Operators should be formatted with a single space on both sides of
            an infix operator. The different value for this setting are ``1``,
            ``0`` or a number greater that ``1``.
        """
        naming_convention_map = {
            'camel': 'camel_case',
            'snake': 'snake_case',
            'hyphen': 'hyphenated_lowercase'
        }
        space_setting_map = {'one_space': 1, 'no_space': 0}
        options = {'BangFormat': {'enabled': True,
                                  'space_before_bang': space_around_bang[0],
                                  'space_after_bang': space_around_bang[1]},
                   'ChainedClasses': {'enabled': not allow_chained_classes},
                   'ColorKeyword': {'enabled': not prefer_color_keywords},
                   'ColorVariable': {'enabled': use_color_variables},
                   'DebugStatement': {'enabled': not allow_debug_statement},
                   'DeclarationOrder': {'enabled': check_declaration_order},
                   'DuplicateProperty': {
                       'enabled': not allow_duplicate_properties,
                       'ignore_consecutive':
                           allow_consecutives_duplicate_property},
                   'ElsePlacement': {'enabled': True,
                                     'style': ('same_line' if else_on_same_line
                                               else '')},
                   'EmptyLineBetweenBlocks': {
                       'enabled': force_empty_line_between_blocks,
                       'ignore_single_line_blocks': True},
                   'EmptyRule': {'enabled': not allow_empty_rules},
                   'HexLength': {'enabled': True,
                                 'style': ('short'
                                           if
                                           use_short_hexadecimal_length_style
                                           else 'long')},
                   'HexNotation': {'enabled': True,
                                   'style': ('lowercase'
                                             if use_lowercase_hexadecimal
                                             else 'uppercase')},
                   'HexValidation': {'enabled': validate_hexadecimal},
                   'IdSelector': {'enabled': not allow_id_selector},
                   'ImportantRule': {'enabled':
                                     not allow_important_rule_in_properties},
                   'Indentation': {'enabled': True,
                                   'allow_non_nested_indentation': False,
                                   'character': ('space'
                                                 if use_spaces else 'tab'),
                                   'width': indent_size},
                   'LeadingZero': {
                                   'enabled': True,
                                   'style': (
                                       'exclude_zero' if exclude_leading_zero
                                       else 'include_zero')
                                  },
                   'MergeableSelector': {'enabled': allow_mergeable_selectors,
                                         'force_nesting': True},
                   'NestingDepth': {'enabled': True,
                                    'max_depth': max_nesting_depth,
                                    'ignore_parent_selectors': False},
                   'NameFormat': {
                       'enabled': True,
                       'allow_leading_underscore': allow_leading_underscore,
                       'function_convention':
                           naming_convention_map.get(
                               function_naming_convention,
                               function_naming_convention),
                       'mixin_convention': naming_convention_map.get(
                           mixin_naming_convention, mixin_naming_convention),
                       'variable_convention':
                           naming_convention_map.get(
                               variable_naming_convention,
                               variable_naming_convention),
                       'placeholder_convention':
                           naming_convention_map.get(
                              placeholder_naming_convention,
                              placeholder_naming_convention)},
                   'PlaceholderInExtend': {
                       'enabled': use_placeholder_selector_in_extend},
                   'PropertyCount': {'enabled': False,
                                     'include_nested': False,
                                     'max_properties': max_properties},
                   'ZeroUnit': {'enabled': not allow_unit_on_zero_values},
                   'UrlFormat': {'enabled': check_ulrs_format},
                   'UrlQuotes': {'enabled': urls_in_quotes},
                   'UnnecessaryMantissa': {
                       'enabled': not allow_unnecessary_mantissa},
                   'UnnecessaryParentReference': {
                       'enabled': not allow_unnecesseary_parent_reference},
                   'TrailingSemicolon': {'enabled': allow_trailing_semicolon},
                   'TrailingWhitespace': {
                       'enabled': not allow_trailing_whitespaces},
                   'ImportPath': {
                       'enabled': check_imports_path,
                       'leading_underscore': allow_filename_leading_underscore,
                       'filename_extension': allow_filename_extension},
                   'LengthVariable': {'enabled': use_length_variables},
                   'PropertySpelling': {
                       'enabled': check_properties_spelling,
                       'extra_properties': extra_properties,
                       'disabled_properties': disabled_properties},
                   'SpaceBetweenParens': {'enabled': True,
                                          'spaces':
                                          spaces_between_parentheses},
                   'SpaceAroundOperator': {
                       'enabled': True,
                       'style': space_setting_map.get(spaces_around_operators,
                                                      'at_least_one_space')}}
        configs = {'linters': options}
        return yaml.dump(configs, default_flow_style=False)



from collections import namedtuple
import pkg_resources
import re

from safety import safety

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange
from coalib.settings.Setting import typed_list


# the safety module expects an object that looks like this
# (not importing it from there because it's in a private-ish location)
Package = namedtuple('Package', ('key', 'version'))


class PySafetyBear(LocalBear):
    """
    Checks if any of your Python dependencies have known security issues.

    Data is taken from pyup.io's vulnerability database hosted at
    https://github.com/pyupio/safety.
    """

    LANGUAGES = {
        'Python Requirements',
        'Python 2 Requirements',
        'Python 3 Requirements',
    }
    AUTHORS = {'Bence Nagy'}
    REQUIREMENTS = {PipRequirement('safety', '0.5.1')}
    AUTHORS_EMAILS = {'bence@underyx.me'}
    LICENSE = 'AGPL'
    CAN_DETECT = {'Security'}

    def run(self, filename, file):
        """
        Checks for vulnerable package versions in requirements files.
        """
        packages = list(
            Package(key=req.key, version=req.specs[0][1])
            for req in self.try_parse_requirements(file)
            if len(req.specs) == 1 and req.specs[0][0] == '=='
        )

        if not packages:
            return

        for vulnerability in safety.check(packages=packages):
            if vulnerability.is_cve:
                message_template = (
                    '{vuln.name}{vuln.spec} is vulnerable to {vuln.cve_id} '
                    'and your project is using {vuln.version}.'
                )
            else:
                message_template = (
                    '{vuln.name}{vuln.spec} is vulnerable and your project is '
                    'using {vuln.version}.'
                )

            # StopIteration should not ever happen so skipping its branch
            line_number, line = next(  # pragma: no branch
                (index, line) for index, line in enumerate(file, start=1)
                if vulnerability.name in line
            )
            version_spec_match = re.search(r'[=<>]+(\S+?)(?:$|\s|#)', line)
            source_range = SourceRange.from_values(
                filename,
                line_number,
                version_spec_match.start(1) + 1,
                line_number,
                version_spec_match.end(1) + 1,
            )

            yield Result(
                self,
                message_template.format(vuln=vulnerability),
                additional_info=vulnerability.description,
                affected_code=(source_range, ),
            )

    @staticmethod
    def try_parse_requirements(lines: typed_list(str)):
        """
        Yields all package requirements parseable from the given lines.

        :param lines: An iterable of lines from a requirements file.
        """
        for line in lines:
            try:
                yield from pkg_resources.parse_requirements(line)
            except pkg_resources.RequirementParseError:
                # unsupported requirement specification
                pass



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.settings.Setting import path


@linter(executable='cmakelint',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+): (?P<message>.*)')
class CMakeLintBear:
    """
    Check CMake code for syntactical or formatting issues.

    For more information consult <https://github.com/richq/cmake-lint>.
    """
    LANGUAGES = {'CMake'}
    REQUIREMENTS = {PipRequirement('cmakelint', '1.3')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         cmakelint_config: path=''):
        """
        :param cmakelint_config: The location of the cmakelintrc config file.
        """
        args = ()
        if cmakelint_config:
            args += ('--config=' + cmakelint_config,)
        return args + (filename,)



import os

from coalib.results.Result import Result
from coalib.bears.GlobalBear import GlobalBear


class PythonPackageInitBear(GlobalBear):
    """
    Looks for missing __init__.py files in directories containing python files.
    """

    LANGUAGES = {'Python', 'Python 3', 'Python 2'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'

    def run(self):
        dirs = {os.path.split(filename)[0]
                for filename in self.file_dict.keys()
                if filename.endswith('.py')}

        missing_inits = {directory for directory in dirs
                         if not os.path.join(directory, '__init__.py')
                         in self.file_dict}

        for missing_init_dir in missing_inits:
            yield Result(self,
                         'Directory "{}" does not contain __init__.py file'
                         .format(os.path.relpath(missing_init_dir,
                                                 self.get_config_dir())))



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='csslint',
        output_format='regex',
        output_regex=r'.+: *(?:line (?P<line>\d+), '
                     r'col (?P<column>\d+), )?(?P<severity>Error|Warning) - '
                     r'(?P<message>.*)')
class CSSLintBear:
    """
    Check code for syntactical or semantical problems that might lead to
    problems or inefficiencies.
    """
    LANGUAGES = {'CSS'}
    REQUIREMENTS = {NpmRequirement('csslint', '1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Code Simplification'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--format=compact', filename



from isort import SortImports

from coalib.bearlib import deprecate_settings
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.settings.Setting import typed_list


class PyImportSortBear(LocalBear):

    LANGUAGES = {'Python', 'Python 3', 'Python 2'}
    REQUIREMENTS = {PipRequirement('isort', '4.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    @deprecate_settings(indent_size='tab_width')
    def run(self, filename, file,
            use_parentheses_in_import: bool=True,
            force_alphabetical_sort_in_import: bool=False,
            force_sort_within_import_sections: bool=True,
            from_first_in_import: bool=False,
            include_trailing_comma_in_import: bool=False,
            combine_star_imports: bool=True,
            combine_as_imports: bool=True,
            lines_after_imports: int=-1,
            order_imports_by_type: bool=False,
            balanced_wrapping_in_imports: bool=False,
            import_heading_localfolder: str='',
            import_heading_firstparty: str='',
            import_heading_thirdparty: str='',
            import_heading_stdlib: str='',
            import_heading_future: str='',
            default_import_section: str='FIRSTPARTY',
            force_grid_wrap_imports: bool=False,
            force_single_line_imports: bool=True,
            sort_imports_by_length: bool=False,
            use_spaces: bool=True,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            forced_separate_imports: typed_list(str)=(),
            isort_multi_line_output: int=4,
            known_first_party_imports: typed_list(str)=(),
            known_third_party_imports: typed_list(str)=(),
            known_standard_library_imports: typed_list(str)=None,
            max_line_length: int=79,
            imports_forced_to_top: typed_list(str)=()):
        """
        Raise issues related to sorting imports, segregating imports into
        various sections, and also adding comments on top of each import
        section based on the configurations provided.

        You can read more about ``isort`` at
        <https://isort.readthedocs.org/en/latest/>.

        :param use_parentheses_in_import:
            True if parenthesis are to be used in import statements.
        :param force_alphabetical_sort_in_import:
            If set, forces all imports to be sorted as a single section,
            instead of within other groups (such as straight vs from).
        :param force_sort_within_import_sections:
            If set, imports will be sorted within there section independent
            to the import_type.
        :param from_first_in_import:
            If set, imports using "from" will be displayed above normal
            (straight) imports.
        :param include_trailing_comma_in_import:
            If set, will automatically add a trailing comma to the end of
            "from" imports. Example: ``from abc import (a, b, c,)``
        :param combine_star_imports:
            If set to true - ensures that if a star import is present,
            nothing else is imported from that namespace.
        :param combine_as_imports:
            If set to true - isort will combine as imports on the same line
            within for import statements.
        :param lines_after_imports:
            Forces a certain number of lines after the imports and before the
            first line of functional code. By default this is set to -1 which
            uses 2 lines if the first line of code is a class or function and
            1 line if it's anything else.
        :param order_imports_by_type:
            If set to true - isort will create separate sections within "from"
            imports for CONSTANTS, Classes, and modules/functions.
        :param balanced_wrapping_in_imports:
            If set to true - for each multi-line import statement isort will
            dynamically change the import length to the one that produces
            the most balanced grid, while staying below the maximum import
            length defined.
        :param import_heading_localfolder:
            A comment to consistently place directly above imports that
            start with '.'.
        :param import_heading_firstparty:
            A comment to consistently place directly above imports from
            the current project.
        :param import_heading_thirdparty:
            A comment to consistently place directly above thirdparty imports.
        :param import_heading_stdlib:
            A comment to consistently place directly above imports from
            the standard library.
        :param import_heading_future:
            A comment to consistently place directly above future imports.
        :param default_import_section:
            The default section to place imports in, if their section can
            not be automatically determined.
        :param force_grid_wrap_imports:
             Force "from" imports to be grid wrapped regardless of line length.
        :param force_single_line_imports:
            If set to true - instead of wrapping multi-line from style imports,
            each import will be forced to display on its own line.
        :param sort_imports_by_length:
            Set to true to sort imports by length instead of alphabetically.
        :param use_spaces:
            True if spaces are to be used instead of tabs.
        :param indent_size:
            Number of spaces per indentation level.
        :param forced_separate_imports:
            A list of modules that you want to appear in their own separate
            section.
        :param isort_multi_line_output:
            An integer that represents how you want imports to be displayed
            by ``isort`` if they're long enough to span multiple lines.
            This value is passed to isort as the ``multi_line_output`` setting.
            Possible values are (0-grid, 1-vertical, 2-hanging, 3-vert-hanging,
            4-vert-grid, 5-vert-grid-grouped)
            A full definition of all possible modes can be found at
            <https://github.com/timothycrosley/isort#multi-line-output-modes>.
        :param known_first_party_imports:
            A list of imports that will be forced to display within the
            standard library category of imports.
        :param known_third_party_imports:
            A list of imports that will be forced to display within the
            third party category of imports.
        :param known_standard_library_imports:
            A list of imports that will be forced to display within the
            first party category of imports.
        :param import_wrap_length:
            An integer that represents the longest line-length you want when
            wrapping. If not set will default to line_length.
        :param imports_forced_to_top:
            Forces a list of imports to the top of their respective section.
            This works well for handling the unfortunate cases of import
            dependencies that occur in many projects.
        :param max_line_length:
            Maximum number of characters for a line.
        """
        isort_settings = dict(
            use_parentheses=use_parentheses_in_import,
            force_alphabetical_sort=force_alphabetical_sort_in_import,
            force_sort_within_sections=force_sort_within_import_sections,
            from_first=from_first_in_import,
            include_trailing_comma=include_trailing_comma_in_import,
            combine_star=combine_star_imports,
            lines_after_imports=lines_after_imports,
            order_by_type=order_imports_by_type,
            balanced_wrapping=balanced_wrapping_in_imports,
            import_heading_localfolder=import_heading_localfolder,
            import_heading_firstparty=import_heading_firstparty,
            import_heading_thirdparty=import_heading_thirdparty,
            import_heading_stdlib=import_heading_stdlib,
            import_heading_future=import_heading_future,
            default_section=default_import_section,
            force_grid_wrap=force_grid_wrap_imports,
            force_single_line=force_single_line_imports,
            length_sort=sort_imports_by_length,
            indent='Tab' if not use_spaces else indent_size,
            forced_separate=forced_separate_imports,
            multi_line_output=isort_multi_line_output,
            known_first_party=known_first_party_imports,
            known_third_party=known_third_party_imports,
            line_length=max_line_length,
            force_to_top=imports_forced_to_top)

        if known_standard_library_imports is not None:
            isort_settings['known_standard_library'] = (
                known_standard_library_imports)

        sort_imports = SortImports(file_contents=''.join(file),
                                   **isort_settings)
        new_file = tuple(sort_imports.output.splitlines(True))

        if new_file != tuple(file):
            diff = Diff.from_string_arrays(file, new_file)
            yield Result(self,
                         'Imports can be sorted.',
                         affected_code=diff.affected_code(filename),
                         diffs={filename: diff})



import autopep8

from coalib.bearlib import deprecate_settings
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.settings.Setting import typed_list


class PEP8Bear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('autopep8', '1.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    @deprecate_settings(indent_size='tab_width')
    def run(self, filename, file,
            max_line_length: int=79,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            pep_ignore: typed_list(str)=(),
            pep_select: typed_list(str)=(),
            local_pep8_config: bool=False):
        """
        Detects and fixes PEP8 incompliant code. This bear will not change
        functionality of the code in any way.

        :param max_line_length:   Maximum number of characters for a line.
        :param indent_size:       Number of spaces per indentation level.
        :param pep_ignore:        A list of errors/warnings to ignore.
        :param pep_select:        A list of errors/warnings to exclusively
                                  apply.
        :param local_pep8_config: Set to true if autopep8 should use a config
                                  file as if run normally from this directory.
        """
        options = {'ignore': pep_ignore,
                   'select': pep_select,
                   'max_line_length': max_line_length,
                   'indent_size': indent_size}

        corrected = autopep8.fix_code(''.join(file),
                                      apply_config=local_pep8_config,
                                      options=options).splitlines(True)

        diffs = Diff.from_string_arrays(file, corrected).split_diff()

        for diff in diffs:
            yield Result(self,
                         'The code does not comply to PEP8.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff})



import nltk
import re
import shutil
import os
from urllib.parse import urlparse

from coalib.bears.GlobalBear import GlobalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coala_utils.ContextManagers import change_directory
from coalib.misc.Shell import run_shell_command
from coalib.results.Result import Result
from coalib.settings.FunctionMetadata import FunctionMetadata
from coalib.settings.Setting import typed_list


class GitCommitBear(GlobalBear):
    LANGUAGES = {'Git'}
    REQUIREMENTS = {PipRequirement('nltk', '3.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/e146c9739ojhr8396wedsvf0d'
    CAN_DETECT = {'Formatting'}
    SUPPORTED_HOST_KEYWORD_REGEX = {
        'github': (r'[Cc]lose[sd]?'
                   r'|[Rr]esolve[sd]?'
                   r'|[Ff]ix(?:e[sd])?'),
        'gitlab': (r'[Cc]los(?:e[sd]?|ing)'
                   r'|[Rr]esolv(?:e[sd]?|ing)'
                   r'|[Ff]ix(?:e[sd]|ing)?')
    }
    CONCATENATION_KEYWORDS = [r',', r'\sand\s']

    _nltk_data_downloaded = False

    def setup_dependencies(self):
        if not self._nltk_data_downloaded and bool(
                self.section.get('shortlog_check_imperative', True)):
            nltk.download([
                'punkt',
                'maxent_treebank_pos_tagger',
                'averaged_perceptron_tagger',
            ])
            type(self)._nltk_data_downloaded = True

    @classmethod
    def check_prerequisites(cls):
        if shutil.which('git') is None:
            return 'git is not installed.'
        else:
            return True

    @classmethod
    def get_shortlog_checks_metadata(cls):
        return FunctionMetadata.from_function(
            cls.check_shortlog,
            omit={'self', 'shortlog'})

    @classmethod
    def get_body_checks_metadata(cls):
        return FunctionMetadata.from_function(
            cls.check_body,
            omit={'self', 'body'})

    @classmethod
    def get_issue_checks_metadata(cls):
        return FunctionMetadata.from_function(
            cls.check_issue_reference,
            omit={'self', 'body'})

    @classmethod
    def get_metadata(cls):
        return FunctionMetadata.merge(
            FunctionMetadata.from_function(
                cls.run,
                omit={'self', 'dependency_results'}),
            cls.get_shortlog_checks_metadata(),
            cls.get_body_checks_metadata(),
            cls.get_issue_checks_metadata())

    @staticmethod
    def get_host_from_remotes():
        """
        Retrieve the first host from the list of git remotes.
        """
        remotes, _ = run_shell_command(
            "git config --get-regex '^remote.*.url$'")

        remotes = [url.split()[-1] for url in remotes.splitlines()]
        if len(remotes) == 0:
            return None

        url = remotes[0]
        if 'git@' in url:
            netloc = re.findall(r'@(\S+):', url)[0]
        else:
            netloc = urlparse(url)[1]
        return netloc.split('.')[0]

    def run(self, allow_empty_commit_message: bool = False, **kwargs):
        """
        Check the current git commit message at HEAD.

        This bear ensures automatically that the shortlog and body do not
        exceed a given line-length and that a newline lies between them.

        :param allow_empty_commit_message: Whether empty commit messages are
                                           allowed or not.
        """
        with change_directory(self.get_config_dir() or os.getcwd()):
            stdout, stderr = run_shell_command('git log -1 --pretty=%B')

        if stderr:
            self.err('git:', repr(stderr))
            return

        stdout = stdout.rstrip('\n')
        pos = stdout.find('\n')
        shortlog = stdout[:pos] if pos != -1 else stdout
        body = stdout[pos+1:] if pos != -1 else ''

        if len(stdout) == 0:
            if not allow_empty_commit_message:
                yield Result(self, 'HEAD commit has no message.')
            return

        yield from self.check_shortlog(
            shortlog,
            **self.get_shortlog_checks_metadata().filter_parameters(kwargs))
        yield from self.check_body(
            body,
            **self.get_body_checks_metadata().filter_parameters(kwargs))
        yield from self.check_issue_reference(
            body,
            **self.get_issue_checks_metadata().filter_parameters(kwargs))

    def check_shortlog(self, shortlog,
                       shortlog_length: int=50,
                       shortlog_regex: str='',
                       shortlog_trailing_period: bool=None,
                       shortlog_imperative_check: bool=True,
                       shortlog_wip_check: bool=True):
        """
        Checks the given shortlog.

        :param shortlog:                 The shortlog message string.
        :param shortlog_length:          The maximum length of the shortlog.
                                         The newline character at end does not
                                         count to the length.
        :param shortlog_regex:           A regex to check the shortlog with.
        :param shortlog_trailing_period: Whether a dot shall be enforced at end
                                         end or not (or ``None`` for "don't
                                         care").
        :param shortlog_wip_check:       Whether a "WIP" in the shortlog text
                                         should yield a result or not.
        """
        diff = len(shortlog) - shortlog_length
        if diff > 0:
            yield Result(self,
                         'Shortlog of the HEAD commit contains {} '
                         'character(s). This is {} character(s) longer than '
                         'the limit ({} > {}).'.format(
                              len(shortlog), diff,
                              len(shortlog), shortlog_length))

        if (shortlog[-1] != '.') == shortlog_trailing_period:
            yield Result(self,
                         'Shortlog of HEAD commit contains no period at end.'
                         if shortlog_trailing_period else
                         'Shortlog of HEAD commit contains a period at end.')

        if shortlog_regex:
            match = re.fullmatch(shortlog_regex, shortlog)
            if not match:
                yield Result(
                    self,
                    'Shortlog of HEAD commit does not match given regex:'
                    ' {regex}'.format(regex=shortlog_regex))

        if shortlog_imperative_check:
            colon_pos = shortlog.find(':')
            shortlog = (shortlog[colon_pos + 1:]
                        if colon_pos != -1
                        else shortlog)
            has_flaws = self.check_imperative(shortlog)
            if has_flaws:
                bad_word = has_flaws[0]
                yield Result(self,
                             "Shortlog of HEAD commit isn't in imperative "
                             "mood! Bad words are '{}'".format(bad_word))
        if shortlog_wip_check:
            if 'wip' in shortlog.lower()[:4]:
                yield Result(
                    self,
                    'This commit seems to be marked as work in progress and '
                    'should not be used in production. Treat carefully.')

    def check_imperative(self, paragraph):
        """
        Check the given sentence/s for Imperatives.

        :param paragraph:
            The input paragraph to be tested.
        :return:
            A list of tuples having 2 elements (invalid word, parts of speech)
            or an empty list if no invalid words are found.
        """
        words = nltk.word_tokenize(nltk.sent_tokenize(paragraph)[0])
        # VBZ : Verb, 3rd person singular present, like 'adds', 'writes'
        #       etc
        # VBD : Verb, Past tense , like 'added', 'wrote' etc
        # VBG : Verb, Present participle, like 'adding', 'writing'
        word, tag = nltk.pos_tag(['I'] + words)[1:2][0]
        if(tag.startswith('VBZ') or
           tag.startswith('VBD') or
           tag.startswith('VBG') or
           word.endswith('ing')):  # Handle special case for VBG
            return (word, tag)
        else:
            return None

    def check_body(self, body,
                   body_line_length: int=72,
                   force_body: bool=False,
                   ignore_length_regex: typed_list(str)=(),
                   body_regex: str=None):
        """
        Checks the given commit body.

        :param body:                The body of the commit message of HEAD.
        :param body_line_length:    The maximum line-length of the body. The
                                    newline character at each line end does not
                                    count to the length.
        :param force_body:          Whether a body shall exist or not.
        :param ignore_length_regex: Lines matching each of the regular
                                    expressions in this list will be ignored.
        :param body_regex:          If provided, checks the presence of regex
                                    in the commit body.
        """
        if len(body) == 0:
            if force_body:
                yield Result(self, 'No commit message body at HEAD.')
            return

        if body[0] != '\n':
            yield Result(self, 'No newline found between shortlog and body at '
                               'HEAD commit. Please add one.')
            return

        if body_regex and not re.fullmatch(body_regex, body.strip()):
            yield Result(self, 'No match found in commit message for the '
                               'regular expression provided: %s' % body_regex)

        body = body.splitlines()
        ignore_regexes = [re.compile(regex) for regex in ignore_length_regex]
        if any((len(line) > body_line_length and
                not any(regex.search(line) for regex in ignore_regexes))
               for line in body[1:]):
            yield Result(self, 'Body of HEAD commit contains too long lines. '
                               'Commit body lines should not exceed {} '
                               'characters.'.format(body_line_length))

    def check_issue_reference(self, body,
                              body_close_issue: bool=False,
                              body_close_issue_full_url: bool=False,
                              body_close_issue_on_last_line: bool=False,
                              body_enforce_issue_reference: bool=False):
        """
        Check for matching issue related references and URLs.

        :param body:
            Body of the commit message of HEAD.
        :param body_close_issue:
            GitHub and GitLab support auto closing issues with
            commit messages. When enabled, this checks for matching keywords
            in the commit body by retrieving host information from git
            configuration. By default, if none of ``body_close_issue_full_url``
            and ``body_close_issue_on_last_line`` are enabled, this checks for
            presence of short references like ``closes #213``.
            Otherwise behaves according to other chosen flags.
            More on keywords follows.
            [GitHub](https://help.github.com/articles/closing-issues-via-commit-messages/)
            [GitLab](https://docs.gitlab.com/ce/user/project/issues/automatic_issue_closing.html)
        :param body_close_issue_full_url:
            Checks the presence of issue close reference with a full URL
            related to some issue. Works along with ``body_close_issue``.
        :param body_close_issue_on_last_line:
            When enabled, checks for issue close reference presence on the
            last line of the commit body. Works along with
            ``body_close_issue``.
        :param body_enforce_issue_reference:
            Whether to enforce presence of issue reference in the body of
            commit message.
        """
        if not body_close_issue:
            return

        host = self.get_host_from_remotes()
        if host not in self.SUPPORTED_HOST_KEYWORD_REGEX:
            return

        if body_close_issue_on_last_line:
            body = body.splitlines()[-1]
            result_message = ('Body of HEAD commit does not contain any {} '
                              'reference in the last line.')
        else:
            result_message = ('Body of HEAD commit does not contain any {} '
                              'reference.')

        if body_close_issue_full_url:
            result_info = 'full issue'
            issue_ref_regex = (
                r'https?://{}\S+/issues/(\S+)'.format(re.escape(host)))
        else:
            result_info = 'issue'
            issue_ref_regex = r'(?:\w+/\w+)?#(\S+)'

        concat_regex = '|'.join(kw for kw in self.CONCATENATION_KEYWORDS)
        compiled_joint_regex = re.compile(
            r'(?:{0})\s+'           # match issue related keywords,
                                    # eg: fix, closes etc.

            r'((?:\S(?!{1}))*\S'    # match links/tags
                                    # eg: fix #123, fix https://github.com

            r'(?:\s*(?:{1})\s*'     # match conjunctions like ',','and'

            r'(?!{0})'              # reject if new keywords appear

            r'(?:\S(?!{1}))*\S)*)'  # match links/tags followed after
                                    # conjunctions if any
            r''.format(
                self.SUPPORTED_HOST_KEYWORD_REGEX[host],
                concat_regex))

        matches = compiled_joint_regex.findall(body)

        if body_enforce_issue_reference and len(matches) == 0:
            yield Result(self, result_message.format(result_info))
            return

        compiled_issue_ref_regex = re.compile(issue_ref_regex)
        compiled_issue_no_regex = re.compile(r'[1-9][0-9]*')
        compiled_concat_regex = re.compile(
            r'\s*(?:{})\s*'.format(concat_regex))

        for match in matches:
            for issue in re.split(compiled_concat_regex, match):
                reference = compiled_issue_ref_regex.fullmatch(issue)
                if not reference:
                    yield Result(self, 'Invalid {} reference: '
                                       '{}'.format(result_info, issue))
                elif not compiled_issue_no_regex.fullmatch(reference.group(1)):
                    yield Result(self, 'Invalid issue number: '
                                       '{}'.format(issue))



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='goimports',
        use_stdin=True,
        output_format='corrected',
        result_message='Imports need to be added/removed.')
class GoImportsBear:
    """
    Adds/Removes imports to Go code for missing imports.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(
        package='golang.org/x/tools/cmd/goimports', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Missing import'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ()



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='hlint')
class HaskellLintBear:
    """
    Check Haskell code for possible problems. This bear can propose patches for
    using alternative functions, simplifying code and removing redundancies.

    See <http://community.haskell.org/~ndm/darcs/hlint/hlint.htm> for more
    information.
    """

    LANGUAGES = {'Haskell'}
    REQUIREMENTS = {DistributionRequirement(apt_get='hlint')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Duplication'}
    CAN_FIX = {'Unused Code', 'Code Simplification'}

    severity_map = {'Error': RESULT_SEVERITY.MAJOR,
                    'Warning': RESULT_SEVERITY.NORMAL,
                    'Suggestion': RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--json', filename

    def process_output(self, output, filename, file):
        output = json.loads(output)

        for issue in output:
            diff = Diff(file)
            from_lines = issue['from'].splitlines()
            to_lines = issue['to'].splitlines()
            assert len(from_lines) == len(to_lines)
            for other_lines in range(1, len(from_lines)):
                assert from_lines[other_lines] == to_lines[other_lines]
            line_nr = issue['startLine']
            line_to_change = file[line_nr-1]
            newline = line_to_change.replace(from_lines[0], to_lines[0])
            diff.change_line(line_nr, line_to_change, newline)

            yield Result.from_values(
                origin=self,
                message=issue['hint'],
                file=filename,
                severity=self.severity_map[issue['severity']],
                line=issue['startLine'],
                column=issue['startColumn'],
                end_line=issue['endLine'],
                end_column=issue['endColumn'],
                diffs={filename: diff})



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement
from coalib.settings.Setting import typed_list


@linter(executable='errcheck',
        output_format='regex',
        output_regex=r'[^:]+:(?P<line>\d+):'
                     r'(?P<column>\d+)\s*(?P<message>.*)',
        result_message='This function call has an unchecked error.')
class GoErrCheckBear:

    """
    Checks the code for all function calls that have unchecked errors.
    GoErrCheckBear runs ``errcheck`` over each file to find such functions.

    For more information on the analysis visit
    <https://github.com/kisielk/errcheck>.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(
        package='github.com/kisielk/errcheck', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/46834'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         ignore: typed_list(str)=[],
                         ignorepkg: typed_list(str)=[],
                         asserts: bool=False,
                         blank: bool=False):
        """
        Bear configuration arguments.

        :param ignore:       Comma-separated list of pairs of the
                             form package:regex. For each package, the regex
                             describes which functions to ignore within that
                             package. The package may be omitted to have the
                             regex apply to all packages.
        :param ignorepkg:    Takes a comma-separated list of package
                             import paths to ignore.
        :param asserts:      Enables checking for ignored type assertion
                             results.
        :param blank:        Enables checking for assignments of errors to
                             the blank identifier.

        """
        args = ()
        if ignore:
            args += ('-ignore', ','.join(part.strip() for part in ignore))
        if ignorepkg:
            args += ('-ignorepkg', ','.join(part.strip()
                                            for part in ignorepkg))
        if blank:
            args += ('-blank',)
        if asserts:
            args += ('-asserts',)
        return args + (filename,)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='vint', output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): (?P<message>.+)')
class VintBear:
    """
    Check vimscript code for possible style problems.

    See <https://github.com/Kuniwak/vint> for more information.
    """

    LANGUAGES = {'VimScript'}
    REQUIREMENTS = {PipRequirement('vim-vint', '0.3.10')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='flawfinder',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+):\s*'
                     r'\[(?P<severity>\d)\]\s*'
                     r'\((?P<origin>[^\)]+)\) (?P<message>.+)',
        severity_map={'1': RESULT_SEVERITY.INFO,
                      '2': RESULT_SEVERITY.INFO,
                      '3': RESULT_SEVERITY.NORMAL,
                      '4': RESULT_SEVERITY.NORMAL,
                      '5': RESULT_SEVERITY.MAJOR},
        prerequisite_check_command=('flawfinder',),
        prerequisite_check_fail_message=('Flawfinder needs to be run with '
                                         'python2. If you have it installed, '
                                         'make sure the first line in the '
                                         'script is '
                                         '"#!/usr/bin/env python2".'))
class CSecurityBear:
    """
    Report possible security weaknesses for C/C++.

    For more information, consult <http://www.dwheeler.com/flawfinder/>.
    """

    LANGUAGES = {'C', 'C++'}
    REQUIREMENTS = {DistributionRequirement('flawfinder')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/7z8ol9mpsgtuo1096c6jk8hi6'
    CAN_DETECT = {'Security', 'Memory Leak', 'Code Simplification'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--columns', '--dataonly', '--quiet', '--singleline', filename



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='proselint',
        output_format='regex',
        output_regex=r'.+?:(?P<line>\d+):(?P<column>\d+): \S* (?P<message>.+)')
class ProseLintBear:
    """
    Lints the file using `proselint <https://github.com/amperser/proselint>`__.
    Works only with English language text.
    """
    LANGUAGES = {'Natural Language'}
    REQUIREMENTS = {PipRequirement('proselint', '0.7.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Spelling', 'Syntax', 'Formatting', 'Grammar'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



import json

from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coala_utils.param_conversion import negate


def bool_or_str(value):
    try:
        return bool(value)
    except ValueError:
        return str(value)


def bool_or_int(value):
    try:
        return bool(value)
    except ValueError:
        return int(value)


@linter(executable='jshint',
        output_format='regex',
        output_regex=r'.+?: line (?P<line>\d+), col (?P<column>\d+), '
                     r'(?P<message>.+) \((?P<severity>[EWI])\d+\)')
class JSHintBear:
    """
    Detect errors and potential problems in JavaScript code and to enforce
    appropriate coding conventions. For example, problems like syntax errors,
    bugs due to implicit type conversion, leaking variables and much more
    can be detected.

    For more information on the analysis visit <http://jshint.com/>
    """

    LANGUAGES = {'JavaScript'}
    REQUIREMENTS = {NpmRequirement('jshint', '2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax', 'Complexity', 'Unused Code'}

    @staticmethod
    @deprecate_settings(es_version='use_es6_syntax',
                        javascript_strictness=(
                            'allow_global_strict',
                            lambda x: 'global' if x else True),
                        cyclomatic_complexity='maxcomplexity',
                        allow_unused_variables=('prohibit_unused', negate),
                        max_parameters='maxparams',
                        allow_missing_semicolon='allow_missing_semicol',
                        allow_this_statements='allow_this_stmt',
                        allow_with_statements='allow_with_stmt',
                        allow_bitwise_operators=('prohibit_bitwise', negate),
                        max_statements='maxstatements',
                        max_depth='maxdepth',
                        allow_comma_operator=('prohibit_comma', negate),
                        allow_non_breaking_whitespace=(
                            'prohibit_non_breaking_whitespace', negate),
                        allow_prototype_overwrite=(
                            'prohibit_prototype_overwrite', negate),
                        allow_type_coercion=('prohibit_type_coercion', negate),
                        allow_future_identifiers=('future_hostile', negate),
                        allow_typeof=('prohibit_typeof', negate),
                        allow_var_statement=(
                            'prohibit_variable_statements', negate),
                        allow_grouping_operator=('prohibit_groups', negate),
                        allow_variable_shadowing='shadow',
                        use_mozilla_extension='using_mozilla',
                        allow_constructor_functions=('prohibit_new', negate),
                        allow_argument_caller_and_callee=(
                            'prohibit_arg', negate),
                        allow_iterator_property=('iterator', negate),
                        allow_filter_in_forin='force_filter_forin')
    def generate_config(filename, file,
                        allow_bitwise_operators: bool=False,
                        allow_prototype_overwrite: bool=False,
                        force_braces: bool=True,
                        allow_type_coercion: bool=False,
                        allow_future_identifiers: bool=True,
                        allow_typeof: bool=True,
                        allow_filter_in_forin: bool=True,
                        allow_funcscope: bool=False,
                        allow_iterator_property: bool=True,
                        allow_argument_caller_and_callee: bool=False,
                        allow_comma_operator: bool=True,
                        allow_non_breaking_whitespace: bool=False,
                        allow_constructor_functions: bool=True,
                        allow_grouping_operator: bool=True,
                        allow_var_statement: bool=True,
                        allow_missing_semicolon: bool=False,
                        allow_debugger: bool=False,
                        allow_assignment_comparisions: bool=False,
                        allow_eval: bool=False,
                        allow_increment: bool=False,
                        allow_proto: bool=False,
                        allow_scripturls: bool=False,
                        allow_singleton: bool=False,
                        allow_this_statements: bool=False,
                        allow_with_statements: bool=False,
                        use_mozilla_extension: bool=False,
                        javascript_strictness: bool_or_str=True,
                        allow_noyield: bool=False,
                        allow_eqnull: bool=False,
                        allow_last_semicolon: bool=False,
                        allow_func_in_loop: bool=False,
                        allow_expr_in_assignments: bool=False,
                        use_es3_array: bool=False,
                        environment_mootools: bool=False,
                        environment_couch: bool=False,
                        environment_jasmine: bool=False,
                        environment_jquery: bool=False,
                        environment_node: bool=False,
                        environment_qunit: bool=False,
                        environment_rhino: bool=False,
                        environment_shelljs: bool=False,
                        environment_prototypejs: bool=False,
                        environment_yui: bool=False,
                        environment_mocha: bool=True,
                        environment_module: bool=False,
                        environment_wsh: bool=False,
                        environment_worker: bool=False,
                        environment_nonstandard: bool=False,
                        environment_browser: bool=True,
                        environment_browserify: bool=False,
                        environment_devel: bool=True,
                        environment_dojo: bool=False,
                        environment_typed: bool=False,
                        environment_phantom: bool=False,
                        max_statements: bool_or_int=False,
                        max_depth: bool_or_int=False,
                        max_parameters: bool_or_int=False,
                        cyclomatic_complexity: bool_or_int=False,
                        allow_variable_shadowing: bool_or_str=False,
                        allow_unused_variables: bool_or_str=False,
                        allow_latedef: bool_or_str=False,
                        es_version: bool_or_int=5,
                        jshint_config: str=''):
        """
        :param allow_bitwise_operators:
            Allows the use of bitwise operators.
        :param allow_prototype_overwrite:
            This options allows overwriting prototypes of native objects such
            as ``Array``.
        :param force_braces:
            This option requires you to always put curly braces around blocks
            in loops and conditionals.
        :param allow_type_coercion:
            This options allows the use of ``==`` and ``!=``.
        :param allow_future_identifiers:
            This option allows the use of identifiers which are defined in
            future versions of JavaScript.
        :param allow_typeof:
            This option enables warnings about invalid ``typeof`` operator
            values.
        :param allow_filter_in_forin:
            This option requires all ``for in`` loops to filter object's items.
        :param allow_iterator_property:
            This option suppresses warnings about the ``__iterator__``
            property.
        :param allow_funcscope:
            This option suppresses warnings about declaring variables inside of
            control structures while accessing them later from outside.
        :param allow_argument_caller_and_callee:
            This option allows the use of ``arguments.caller`` and
            ``arguments.callee``.
        :param allow_comma_operator:
            This option allows the use of the comma operator.
        :param allow_non_breaking_whitespace:
            Allows "non-breaking whitespace characters".
        :param allow_constructor_functions:
            Allows the use of constructor functions.
        :param allow_grouping_operator:
            This option allows the use of the grouping operator when it is
            not strictly required.
        :param allow_var_statement:
            Allows the use of the ``var`` statement while declaring a variable.
            Should use ``let`` or ``const`` while it is set to ``False``.
        :param allow_missing_semicolon:
            This option suppresses warnings about missing semicolons.
        :param allow_debugger:
            This option suppresses warnings about the ``debugger`` statements.
        :param allow_assignment_comparisions:
            This option suppresses warnings about the use of assignments in
            cases where comparisons are expected.
        :param allow_eval:
            This options suppresses warnings about the use of ``eval``
            function.
        :param allow_increment:
            This option suppresses warnings about the use of unary increment
            and decrement operators.
        :param allow_proto:
            This option suppresses warnings about the ``__proto__`` property.
        :param allow_scripturls:
            This option suppresses warnings about the use of script-targeted
            URLs.
        :param allow_singleton:
            This option suppresses warnings about constructions like
            ``new function () { ... }`` and ``new Object;`` sometimes used to
            produce singletons.
        :param allow_this_statements:
            This option suppresses warnings about possible strict violations
            when the code is running in strict mode and ``this`` is used in a
            non-constructor function.
        :param allow_with_statements:
            This option suppresses warnings about the use of the ``with``
            statement.
        :param use_mozilla_extension:
            This options tells JSHint that your code uses Mozilla JavaScript
            extensions.
        :param javascript_strictness:
            Determines what sort of strictness to use in the JavaScript code.
            The possible options are:

            - "global" - there must be a ``"use strict";`` at global level
            - "implied" - lint the code as if there is a ``"use strict";``
            - "False" - disable warnings about strict mode
            - "True" - there must be a ``"use strict";`` at function level
        :param allow_noyield:
            This option suppresses warnings about generator functions with no
            ``yield`` statement in them.
        :param allow_eqnull:
            This option suppresses warnings about ``== null`` comparisons.
        :param allow_last_semicolon:
            This option suppresses warnings about missing semicolons for the
            last statement.
        :param allow_func_in_loop:
            This option suppresses warnings about functions inside of loops.
        :param allow_expr_in_assignments:
            This option suppresses warnings about the use of expressions where
            normally assignments or function calls are expected.
        :param use_es3_array:
            This option tells JSHintBear ES3 array elision elements, or empty
            elements are used.
        :param environment_mootools:
            This option defines globals exposed by the Mootools.
        :param environment_couch:
            This option defines globals exposed by CouchDB.
        :param environment_jasmine:
            This option defines globals exposed by Jasmine.
        :param environment_jquery:
            This option defines globals exposed by Jquery.
        :param environment_node:
            This option defines globals exposed by Node.
        :param environment_qunit:
            This option defines globals exposed by Qunit.
        :param environment_rhino:
            This option defines globals exposed when the code is running inside
            rhino runtime environment.
        :param environment_shelljs:
            This option defines globals exposed by the ShellJS.
        :param environment_prototypejs:
            This option defines globals exposed by the Prototype.
        :param environment_yui:
            This option defines globals exposed by the YUI JavaScript
            Framework.
        :param environment_mocha:
            This option defines globals exposed by the "BDD" and "TDD" UIs of
            the Mocha unit testing framework.
        :param environment_module:
            This option informs JSHintBear that the input code describes an
            ECMAScript 6 module.
        :param environment_wsh:
            This option defines globals available when the code is running as a
            script for the Windows Script Host.
        :param environment_worker:
            This option defines globals available when the code is running
            inside of a Web Worker.
        :param environment_nonstandard:
            This option defines non- standard but widely adopted globals such
            as ``escape`` and ``unescape``.
        :param environment_browser:
            This option defines globals exposed by modern browsers.
        :param environment_browserify:
            This option defines globals available when using the Browserify.
        :param environment_devel:
            This option defines globals that are usually used for debugging:
            ``console``, ``alert``, etc.
        :param environment_dojo:
            This option defines globals exposed by the Dojo Toolkit.
        :param environment_typed:
            This option defines globals for typed array constructors.
        :param environment_phantom:
            This option defines globals available when your core is running
            inside of the PhantomJS runtime environment.
        :param max_statements:
            Maximum number of statements allowed per function.
        :param max_depth:
            This option lets you control how nested do you want your blocks to
            be.
        :param max_parameters:
            Maximum number of parameters allowed per function.
        :param cyclomatic_complexity:
            Maximum cyclomatic complexity in the code.
        :param allow_variable_shadowing:
            This option suppresses warnings about variable shadowing i.e.
            declaring a variable that had been already declared somewhere in
            the outer scope.

            - "inner" - check for variables defined in the same scope only
            - "outer" - check for variables defined in outer scopes as well
            - False - same as inner
            - True  - allow variable shadowing
        :param allow_unused_variables:
            Allows when variables are defined but never used. This can be set
            to ""vars"" to only check for variables, not function parameters,
            or ""strict"" to check all variables and parameters.
        :param allow_latedef:
            This option allows the use of a variable before it was defined.
            Setting this option to "nofunc" will allow function declarations to
            be ignored.
        :param es_version:
            This option is used to specify the ECMAScript version to which the
            code must adhere to.
        """
        # Assume that when es_version is bool, it is intended for the
        # deprecated use_es6_version
        if es_version is True:
            es_version = 6
        elif es_version is False:
            es_version = 5
        if not jshint_config:
            options = {'bitwise': not allow_bitwise_operators,
                       'freeze': not allow_prototype_overwrite,
                       'curly': force_braces,
                       'eqeqeq': not allow_type_coercion,
                       'futurehostile': not allow_future_identifiers,
                       'notypeof': not allow_typeof,
                       'forin': allow_filter_in_forin,
                       'funcscope': allow_funcscope,
                       'iterator': not allow_iterator_property,
                       'noarg': not allow_argument_caller_and_callee,
                       'nocomma': not allow_comma_operator,
                       'nonbsp': not allow_non_breaking_whitespace,
                       'nonew': not allow_constructor_functions,
                       'undef': True,
                       'singleGroups': not allow_grouping_operator,
                       'varstmt': not allow_var_statement,
                       'asi': allow_missing_semicolon,
                       'debug': allow_debugger,
                       'boss': allow_assignment_comparisions,
                       'evil': allow_eval,
                       'strict': javascript_strictness,
                       'plusplus': allow_increment,
                       'proto': allow_proto,
                       'scripturl': allow_scripturls,
                       'supernew': allow_singleton,
                       'validthis': allow_this_statements,
                       'withstmt': allow_with_statements,
                       'moz': use_mozilla_extension,
                       'noyield': allow_noyield,
                       'eqnull': allow_eqnull,
                       'lastsemic': allow_last_semicolon,
                       'loopfunc': allow_func_in_loop,
                       'expr': allow_expr_in_assignments,
                       'elision': use_es3_array,
                       'mootools': environment_mootools,
                       'couch': environment_couch,
                       'jasmine': environment_jasmine,
                       'jquery': environment_jquery,
                       'node': environment_node,
                       'qunit': environment_qunit,
                       'rhino': environment_rhino,
                       'shelljs': environment_shelljs,
                       'prototypejs': environment_prototypejs,
                       'yui': environment_yui,
                       'mocha': environment_mocha,
                       'module': environment_module,
                       'wsh': environment_wsh,
                       'worker': environment_worker,
                       'nonstandard': environment_nonstandard,
                       'browser': environment_browser,
                       'browserify': environment_browserify,
                       'devel': environment_devel,
                       'dojo': environment_dojo,
                       'typed': environment_typed,
                       'phantom': environment_phantom,
                       'maxerr': 99999,
                       'maxcomplexity': cyclomatic_complexity,
                       'maxdepth': max_depth,
                       'maxparams': max_parameters,
                       'maxstatements': max_statements,
                       'shadow': allow_variable_shadowing,
                       'unused': not allow_unused_variables,
                       'latedef': allow_latedef,
                       'esversion': es_version}

            return json.dumps(options)
        else:
            return None

    @staticmethod
    def create_arguments(filename, file, config_file, jshint_config: str=''):
        """
        :param jshint_config:
            The location of the jshintrc config file. If this option is present
            all the above options are not used. Instead the .jshintrc file is
            used as the configuration file.
        """
        args = ('--verbose', filename, '--config')
        if jshint_config:
            args += (jshint_config,)
        else:
            args += (config_file,)
        return args



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.AnyOneOfRequirements import (
    AnyOneOfRequirements)
from dependency_management.requirements.CondaRequirement import (
    CondaRequirement)
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='ruby',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.+?:(?P<line>\d+): (?P<message>.*?'
                     r'(?P<severity>error|warning)[,:] \S+)\s?'
                     r'(?:\S+\s(?P<column>.*?)\^)?')
class RubySyntaxBear:
    """
    Checks the code with ``ruby -wc`` on each file separately.
    """
    LANGUAGES = {'Ruby'}
    REQUIREMENTS = {
        AnyOneOfRequirements(
            [DistributionRequirement('ruby'),
             CondaRequirement('ruby', '2.2.3', 'bioconda'),
             ],
        ),
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '-wc', filename



import re
from shutil import which

from coalib.bears.GlobalBear import GlobalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.misc.Shell import run_shell_command
from coalib.results.Result import Result


class VultureBear(GlobalBear):
    LANGUAGES = {'Python', 'Python 3'}
    REQUIREMENTS = {PipRequirement('vulture', '0.10.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/82256'
    CAN_DETECT = {'Unused Code'}

    EXECUTABLE = 'vulture'
    OUTPUT_REGEX = re.compile(
        r'(?P<filename>.*):(?P<line>.*):\s*(?P<message>.*)')

    @classmethod
    def check_prerequisites(cls):
        return ('Vulture is missing. Make sure to install it using '
                '`pip3 install vulture`.'
                if which('vulture') is None else True)

    def run(self):
        """
        Check Python code for unused variables and functions using `vulture`.

        See <https://bitbucket.org/jendrikseipp/vulture> for more information.
        """
        stdout_output, _ = run_shell_command(
            (self.EXECUTABLE,) +
            tuple(filename for filename in self.file_dict.keys()),
            cwd=self.get_config_dir())

        for match in re.finditer(self.OUTPUT_REGEX, stdout_output):
            groups = match.groupdict()
            yield Result.from_values(origin=self,
                                     message=groups['message'],
                                     file=groups['filename'],
                                     line=int(groups['line']))



from coala_utils.string_processing.Core import escape
from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from dependency_management.requirements.AnyOneOfRequirements import (
    AnyOneOfRequirements)
from dependency_management.requirements.RscriptRequirement import (
    RscriptRequirement)
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


def _map_to_r_bool(py_bool):
    return 'TRUE' if py_bool else 'FALSE'


@linter(executable='Rscript',
        output_format='corrected',
        prerequisite_check_command=('Rscript', '-e', 'library(formatR)'),
        prerequisite_check_fail_message='Please install formatR for this bear '
                                        'to work.')
class FormatRBear:
    """
    Check and correct formatting of R Code using known formatR utility.
    """
    LANGUAGES = {'R'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    REQUIREMENTS = {
        AnyOneOfRequirements(
            [DistributionRequirement(
                apt_get='r-cran-formatr',
                zypper='R-formatR',
             ),
             RscriptRequirement('formatR'),
             ]
        )
    }
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/0y0oxtak18v492jdyfqwpw1n4'
    CAN_FIX = {'Formatting'}

    @staticmethod
    @deprecate_settings(indent_size='tab_width')
    def create_arguments(filename, file, config_file,
                         r_keep_comments: bool=True,
                         r_keep_blank_lines: bool=True,
                         r_braces_on_next_line: bool=None,
                         r_use_arrows: bool=None,
                         indent_size:
                         int=SpacingHelper.DEFAULT_TAB_WIDTH,
                         r_max_expression_length: int=0):
        """
        :param r_keep_comments:
            Determines whether comments are kept or not.
        :param r_keep_blank_lines:
            Determines whether blank lines are kept or not.
        :param r_braces_on_next_line:
            Determines whether a brace should be placed on the next line.

            Example:
            If ``True``::

                if (...) {

            changes to::

                if (...)
                {

            If ``False`` the brace is placed on the same line.
        :param r_use_arrows:
            Determines whether the assignment operator ``=`` should be replaced
            by an arrow ``<-`` or not.

            Example: If  ``True``, ``x = 1`` changes to ``x <- 1``.
        :param indent_size:
            Number of spaces per indentation level.
        :param r_max_expression_length:
            Maximum number of characters for an expression.

            Example: If ``20`` then::

                1 + 1 + 1 + 1 + 1 + 1 + 1

            changes to::

                1 + 1 + 1 + 1 + 1 + 1 +
                    1
        """
        options = {'source="' + escape(filename, '"\\') + '"',
                   'blank=' + _map_to_r_bool(r_keep_blank_lines),
                   'comment=' + _map_to_r_bool(r_keep_comments),
                   'indent=' + str(indent_size)}
        if r_max_expression_length:
            options.add('width.cutoff=' + str(r_max_expression_length))
        if r_braces_on_next_line is not None:
            options.add('brace.newline=' +
                        _map_to_r_bool(r_braces_on_next_line))
        if r_use_arrows is not None:
            options.add('arrow=' + _map_to_r_bool(r_use_arrows))

        rcode = 'library(formatR);formatR::tidy_source({})'.format(
            ','.join(options))
        return '-e', rcode



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='infer',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+): (?P<severity>error|warning): '
                     r'(?P<message>.*)')
class InferBear:
    """
    Checks the code with ``infer``.
    """
    LANGUAGES = {'Java'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/1g2k0la7xo5az9t8f1v5zy66q'
    CAN_DETECT = {'Security'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '-npb', '--', 'javac', filename



from yapf.yapflib.yapf_api import FormatCode

from coalib.bearlib import deprecate_settings
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coala_utils.ContextManagers import prepare_file
from coalib.results.Result import Result
from coalib.results.Diff import Diff


class YapfBear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    AUTHORS = {'The coala developers'}
    REQUIREMENTS = {PipRequirement('yapf', '0.14.0')}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}
    ASCIINEMA_URL = 'https://asciinema.org/a/89021'

    @deprecate_settings(indent_size='tab_width')
    def run(self, filename, file,
            max_line_length: int=79,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            allow_multiline_lambdas: bool=False,
            blank_line_before_nested_class_or_def: bool=False,
            continuation_tab_width: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            dedent_closing_brackets: bool=False,
            indent_dictionary_value: bool=False,
            coalesce_brackets: bool=False,
            join_multiple_lines: bool=True,
            spaces_around_power_operator: bool=True,
            spaces_before_comment: int=2,
            space_between_ending_comma_and_closing_bracket: bool=True,
            split_arguments_when_comma_terminated: bool=False,
            split_before_bitwise_operator: bool=False,
            split_before_first_argument: bool=False,
            split_before_logical_operator: bool=False,
            split_before_named_assigns: bool=True,
            use_spaces: bool=True,
            based_on_style: str='pep8',
            prefer_line_break_after_opening_bracket: bool=True):
        """
        Check and correct formatting of Python code using ``yapf`` utility.

        See <https://github.com/google/yapf> for more information.

        :param max_line_length:
            Maximum number of characters for a line.
        :param indent_size:
            Number of spaces per indentation level.
        :param allow_multiline_lambdas:
            Allows lambdas to be formatted on more than one line.
        :param blank_line_before_nested_class_or_def:
            Inserts a blank line before a ``def`` or ``class`` immediately
            nested within another ``def`` or ``class``.
        :param continuation_tab_width:
            Indent width used for line continuations.
        :param dedent_closing_brackets:
            Puts closing brackets on a separate line, dedented, if the
            bracketed expression can't fit in a single line. Applies to all
            kinds of brackets, including function definitions and calls.
        :param indent_dictionary_value:
            Indents the dictionary value if it cannot fit on the same line as
            the dictionary key.
        :param coalesce_brackets:
            Prevents splitting consecutive brackets. Only relevant when
            ``dedent_closing_brackets`` is set.
            Example:
            If ``True``::

                call_func_that_takes_a_dict(
                    {
                        'key1': 'value1',
                        'key2': 'value2',
                    }
                )

            would reformat to::

                call_func_that_takes_a_dict({
                    'key1': 'value1',
                    'key2': 'value2',
                })

        :param join_multiple_lines:
            Joins short lines into one line.
        :param spaces_around_power_operator:
            Set to ``True`` to prefer using spaces around ``**``.
        :param spaces_before_comment:
            The number of spaces required before a trailing comment.
        :param space_between_ending_comma_and_closing_bracket:
            Inserts a space between the ending comma and closing bracket of a
            list, etc.
        :param split_arguments_when_comma_terminated:
            Splits before arguments if the argument list is terminated by a
            comma.
        :param split_before_bitwise_operator:
            Set to ``True`` to prefer splitting before ``&``, ``|`` or ``^``
            rather than after.
        :param split_before_first_argument:
            If an argument / parameter list is going to be split, then split
            before the first argument.
        :param split_before_logical_operator:
            Set to ``True`` to prefer splitting before ``and`` or ``or`` rather
            than after.
        :param split_before_named_assigns:
            Splits named assignments into individual lines.
        :param use_spaces:
            Uses spaces for indentation.
        :param based_on_style:
            The formatting style to be used as reference.
        :param prefer_line_break_after_opening_bracket:
            If True, splitting right after a open bracket will not be
            preferred.
        """
        if not file:
            # Yapf cannot handle zero-byte files well, and adds a redundent
            # newline into the file. To avoid this, we don't parse zero-byte
            # files as they cannot have anything to format either.
            return

        options = """
[style]
indent_width = {indent_size}
column_limit = {max_line_length}
allow_multiline_lambdas = {allow_multiline_lambdas}
continuation_indent_width = {continuation_tab_width}
dedent_closing_brackets = {dedent_closing_brackets}
indent_dictionary_value = {indent_dictionary_value}
join_multiple_lines = {join_multiple_lines}
spaces_around_power_operator = {spaces_around_power_operator}
spaces_before_comment = {spaces_before_comment}
coalesce_brackets = {coalesce_brackets}
split_before_bitwise_operator = {split_before_bitwise_operator}
split_before_first_argument = {split_before_first_argument}
split_before_logical_operator = {split_before_logical_operator}
split_before_named_assigns = {split_before_named_assigns}
based_on_style = {based_on_style}
blank_line_before_nested_class_or_def = {blank_line_before_nested_class_or_def}
split_arguments_when_comma_terminated = {split_arguments_when_comma_terminated}
space_between_ending_comma_and_closing_bracket= \
{space_between_ending_comma_and_closing_bracket}
"""
        options += 'use_tabs = ' + str(not use_spaces) + '\n'
        options += ('split_penalty_after_opening_bracket = ' +
                    ('30' if prefer_line_break_after_opening_bracket
                     else '0') + '\n')
        options = options.format(**locals())

        try:
            with prepare_file(options.splitlines(keepends=True),
                              None) as (file_, fname):
                corrected = FormatCode(
                    ''.join(file), style_config=fname)[0].splitlines(True)
        except SyntaxError as err:
            if isinstance(err, IndentationError):
                error_type = 'indentation errors (' + err.args[0] + ')'
            else:
                error_type = 'syntax errors'
            yield Result.from_values(
                self,
                'The code cannot be parsed due to {0}.'.format(error_type),
                filename, line=err.lineno, column=err.offset)
            return
        diffs = Diff.from_string_arrays(file, corrected).split_diff()
        for diff in diffs:
            yield Result(self,
                         'The code does not comply with the settings '
                         'provided.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff})



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='gofmt',
        use_stdin=True,
        output_format='corrected',
        result_message='Formatting can be improved.')
class GofmtBear:
    """
    Suggest better formatting options in Go code. Basic checks like alignment,
    indentation, and redundant parentheses are provided.

    This is done using the ``gofmt`` utility. For more information visit
    <https://golang.org/cmd/gofmt/>.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(package='golang.org/cmd/gofmt', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}
    ASCIINEMA_URL = 'https://asciinema.org/a/94812'

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ()



import os.path
import pyroma

from coalib.bears.GlobalBear import GlobalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Result import Result


class PyromaBear(GlobalBear):
    LANGUAGES = {'Python', 'Python 3'}
    REQUIREMENTS = {PipRequirement('pyroma', '2.2.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'

    def run(self):
        """
        Checks for Python packaging best practices using `pyroma`.

        Pyroma rhymes with aroma, and is a product aimed at giving a rating of
        how well a Python project complies with the best practices of the
        Python packaging ecosystem, primarily PyPI, pip, Distribute etc,
        as well as a list of issues that could be improved.

        See <https://bitbucket.org/regebro/pyroma/> for more information.
        """

        setup_files = [setup_file for setup_file in self.file_dict
                       if os.path.basename(setup_file) == 'setup.py']

        if not setup_files:
            yield Result(self, 'Your package does'
                         ' not contain a setup file.')
        else:
            for setup_file in setup_files:
                data = pyroma.projectdata.get_data(os.path.dirname(setup_file))
                rating = pyroma.ratings.rate(data)
                messages = rating[1]

                for message in messages:
                    yield Result.from_values(origin=self,
                                             message=message,
                                             file=setup_file)



import json
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='stylelint',
        output_format='regex',
        output_regex=r'\s*(?P<filename>.+)\s*(?P<line>\d+):(?P<column>\d+)\s*'
                     r'(\D)\s*(?P<message>.+)',
        config_suffix='.json')
class StyleLintBear:
    """
    Checks the code with stylelint. This will run stylelint over each file
    separately.

    Detect errors and potential problems in CSS code and to enforce
    appropriate coding conventions. For example, problems like syntax errors,
    invalid color codes etc can be detected.

    For more information on the analysis visit <http://stylelint.io/>
    """
    LANGUAGES = {'CSS', 'SCSS'}
    REQUIREMENTS = {NpmRequirement('stylelint', '7')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Unused Code', 'Formatting'}

    @staticmethod
    def generate_config(filename, file):
        # Use standard stylelint rules
        rules = {
            'at-rule-name-case': 'lower',
            'at-rule-name-space-after': 'always-single-line',
            'at-rule-semicolon-newline-after': 'always',
            'block-closing-brace-empty-line-before': 'never',
            'block-closing-brace-newline-after': 'always',
            'block-closing-brace-newline-before': 'always-multi-line',
            'block-closing-brace-space-before': 'always-single-line',
            'block-no-empty': True,
            'block-opening-brace-newline-after': 'always-multi-line',
            'block-opening-brace-space-after': 'always-single-line',
            'block-opening-brace-space-before': 'always',
            'color-hex-case': 'lower',
            'color-hex-length': 'short',
            'color-no-invalid-hex': True,
            'comment-no-empty': True,
            'comment-whitespace-inside': 'always',
            'declaration-bang-space-after': 'never',
            'declaration-bang-space-before': 'always',
            'declaration-block-no-redundant-longhand-properties': True,
            'declaration-block-no-shorthand-property-overrides': True,
            'declaration-block-semicolon-newline-after': 'always-multi-line',
            'declaration-block-semicolon-space-after': 'always-single-line',
            'declaration-block-semicolon-space-before': 'never',
            'declaration-block-single-line-max-declarations': 1,
            'declaration-block-trailing-semicolon': 'always',
            'declaration-colon-newline-after': 'always-multi-line',
            'declaration-colon-space-after': 'always-single-line',
            'declaration-colon-space-before': 'never',
            'font-family-no-duplicate-names': True,
            'function-calc-no-unspaced-operator': True,
            'function-comma-newline-after': 'always-multi-line',
            'function-comma-space-after': 'always-single-line',
            'function-comma-space-before': 'never',
            'function-linear-gradient-no-nonstandard-direction': True,
            'function-max-empty-lines': 0,
            'function-name-case': 'lower',
            'function-parentheses-newline-inside': 'always-multi-line',
            'function-parentheses-space-inside': 'never-single-line',
            'function-whitespace-after': 'always',
            'indentation': 2,
            'keyframe-declaration-no-important': True,
            'length-zero-no-unit': True,
            'max-empty-lines': 1,
            'media-feature-colon-space-after': 'always',
            'media-feature-colon-space-before': 'never',
            'media-feature-name-case': 'lower',
            'media-feature-name-no-unknown': True,
            'media-feature-parentheses-space-inside': 'never',
            'media-feature-range-operator-space-after': 'always',
            'media-feature-range-operator-space-before': 'always',
            'media-query-list-comma-newline-after': 'always-multi-line',
            'media-query-list-comma-space-after': 'always-single-line',
            'media-query-list-comma-space-before': 'never',
            'no-empty-source': True,
            'no-eol-whitespace': True,
            'no-extra-semicolons': True,
            'no-invalid-double-slash-comments': True,
            'no-missing-end-of-source-newline': True,
            'number-leading-zero': 'always',
            'number-no-trailing-zeros': True,
            'property-case': 'lower',
            'property-no-unknown': True,
            'selector-attribute-brackets-space-inside': 'never',
            'selector-attribute-operator-space-after': 'never',
            'selector-attribute-operator-space-before': 'never',
            'selector-combinator-space-after': 'always',
            'selector-combinator-space-before': 'always',
            'selector-descendant-combinator-no-non-space': True,
            'selector-list-comma-newline-after': 'always',
            'selector-list-comma-space-before': 'never',
            'selector-max-empty-lines': 0,
            'selector-pseudo-class-case': 'lower',
            'selector-pseudo-class-no-unknown': True,
            'selector-pseudo-class-parentheses-space-inside': 'never',
            'selector-pseudo-element-case': 'lower',
            'selector-pseudo-element-colon-notation': 'double',
            'selector-pseudo-element-no-unknown': True,
            'selector-type-case': 'lower',
            'selector-type-no-unknown': True,
            'shorthand-property-no-redundant-values': True,
            'string-no-newline': True,
            'unit-case': 'lower',
            'unit-no-unknown': True,
            'value-list-comma-newline-after': 'always-multi-line',
            'value-list-comma-space-after': 'always-single-line',
            'value-list-comma-space-before': 'never',
            'value-list-max-empty-lines': 0
        }

        rules['at-rule-empty-line-before'] = [
             'always',
             {'except': ['blockless-after-same-name-blockless',
                         'first-nested'],
              'ignore': ['after-comment']}
        ]

        rules['comment-empty-line-before'] = [
         'always',
         {'except': ['first-nested'], 'ignore': ['stylelint-commands']}
        ]

        rules['custom-property-empty-line-before'] = [
             'always',
             {'except': ['after-custom-property', 'first-nested'],
              'ignore': ['after-comment', 'inside-single-line-block']}
        ]

        rules['declaration-block-no-duplicate-properties'] = [
             True,
             {'ignore': ['consecutive-duplicates-with-different-values']}
        ]

        rules['declaration-empty-line-before'] = [
             'always',
             {'except': ['after-declaration', 'first-nested'],
              'ignore': ['after-comment', 'inside-single-line-block']}
        ]

        rules['rule-nested-empty-line-before'] = [
         'always-multi-line',
         {'except': ['first-nested'], 'ignore': ['after-comment']}
        ]

        rules['rule-non-nested-empty-line-before'] = [
         'always-multi-line',
         {'ignore': ['after-comment']}
        ]

        default_config = {'rules': rules}
        return json.dumps(default_config)

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename, '--config=' + config_file



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.settings.Setting import typed_list


@linter(executable='bandit')
class BanditBear:
    """
    Performs security analysis on Python source code, utilizing the ``ast``
    module from the Python standard library.
    """
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('bandit', '1.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Security'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         bandit_skipped_tests: typed_list(str)=
                         ('B105', 'B106', 'B107', 'B404', 'B603', 'B606',
                          'B607')):
        """
        :param bandit_skipped_tests:
            The IDs of the tests ``bandit`` shall not perform. You can get
            information about the available builtin codes at
            https://github.com/openstack/bandit#usage.
        """
        args = (filename, '-f', 'json')

        if bandit_skipped_tests:
            args += ('-s', ','.join(bandit_skipped_tests))

        return args

    severity_map = {'HIGH': RESULT_SEVERITY.MAJOR,
                    'MEDIUM': RESULT_SEVERITY.NORMAL,
                    'LOW': RESULT_SEVERITY.INFO}

    confidence_map = {'HIGH': 90,
                      'MEDIUM': 70,
                      'LOW': 50}

    def process_output(self, output, filename, file):
        output = json.loads(output)

        for error in output['errors']:
            yield Result.from_values(
                origin=self,
                file=filename,
                severity=RESULT_SEVERITY.MAJOR,
                message=error['reason'])

        for issue in output['results']:
            yield Result.from_values(
                origin=issue['test_id'],
                file=filename,
                message=issue['issue_text'],
                severity=self.severity_map[issue['issue_severity']],
                confidence=self.confidence_map[issue['issue_confidence']],
                line=issue['line_range'][0],
                end_line=issue['line_range'][-1])



import json
from collections import OrderedDict
from re import match

from coala_utils.param_conversion import negate
from coalib.bearlib import deprecate_settings
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from coalib.misc.Compatibility import JSONDecodeError
from coalib.results.Diff import Diff
from coalib.results.Result import Result


class JSONFormatBear(LocalBear):

    LANGUAGES = {'JSON'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    ASCIINEMA_URL = 'https://asciinema.org/a/6vxc7076tnf996zanpdfwojwu'
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @deprecate_settings(indent_size='tab_width',
                        escape_unicode=('keep_unicode', negate))
    def run(self, filename, file,
            json_sort: bool=False,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            escape_unicode: bool=True):
        """
        Raises issues for any deviations from the pretty-printed JSON.

        :param json_sort:      Whether or not keys should be sorted.
        :param indent_size:    Number of spaces per indentation level.
        :param escape_unicode: Whether or not to escape unicode values using
                               ASCII.
        """
        # Output a meaningful message if empty file given as input
        if len(file) == 0:
            yield Result.from_values(self,
                                     'This file is empty.',
                                     file=filename)
            return

        try:
            json_content = json.loads(''.join(file),
                                      object_pairs_hook=OrderedDict)
        except JSONDecodeError as err:
            err_content = match(r'(.*): line (\d+) column (\d+)', str(err))
            yield Result.from_values(
                self,
                'This file does not contain parsable JSON. ' +
                err_content.group(1) + '.',
                file=filename,
                line=int(err_content.group(2)),
                column=int(err_content.group(3)))
            return

        corrected = json.dumps(json_content,
                               sort_keys=json_sort,
                               indent=indent_size,
                               ensure_ascii=escape_unicode
                               ).splitlines(True)
        # Because of a bug in several python versions we have to correct
        # whitespace here.
        corrected = tuple(line.rstrip(' \n') + '\n' for line in corrected)
        diff = Diff.from_string_arrays(file, corrected)

        if len(diff) > 0:
            yield Result(self,
                         'This file can be reformatted by sorting keys and '
                         'following indentation.',
                         affected_code=tuple(d.range(filename)
                                             for d in diff.split_diff()),
                         diffs={filename: diff})



from clang.cindex import Index, TranslationUnit

from bears.c_languages.ClangBear import clang_available, ClangBear
from coalib.bears.GlobalBear import GlobalBear


class ClangASTPrintBear(GlobalBear):
    check_prerequisites = classmethod(clang_available)
    LANGUAGES = ClangBear.LANGUAGES
    REQUIREMENTS = ClangBear.REQUIREMENTS

    def print_node(self, cursor, filename, before='', spec_before=''):
        '''
        Prints this node and all child nodes recursively in the style of:

        ::

            -node
            |-child
            `-another child
            |- child of child
            `- last child of child

        :param cursor:      The node to print. (Clang cursor.)
        :param before:      What to print before the node.
        :param spec_before: What to print before this node but to replace with
                            spaces for child nodes.
        '''
        file = cursor.location.file

        if file is not None and file.name == filename:
            self.debug(
                before + spec_before + '-' + str(cursor.displayname),
                str(cursor.kind),
                'Lines',
                str(cursor.extent.start.line) + '-' +
                str(cursor.extent.end.line),
                '(' + ' '.join(str(token.spelling)
                               for token in cursor.get_tokens()) + ')')

        children = list(cursor.get_children())

        if len(children) > 0:
            for child in children[:-1]:
                self.print_node(child,
                                filename,
                                before + len(spec_before)*' ' + '|')

            self.print_node(children[-1],
                            filename,
                            before + len(spec_before)*' ',
                            '`')

    def run(self):
        """
        This bear is meant for debugging purposes relating to clang. It just
        prints out the whole AST for a file to the DEBUG channel.
        """
        for filename, file in sorted(self.file_dict.items()):
            root = Index.create().parse(
                filename,
                options=TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD
                ).cursor

            self.print_node(root, filename)



from shutil import which

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='elm-format',
        use_stderr=True,
        use_stdout=False,
        output_format='regex',
        output_regex='(?P<line>\d+)│(?P<message>[\s\S]*)')
class ElmLintBear:
    """
    This bear formats the Elm source code according to a standard set of rules.

    See <https://github.com/avh4/elm-format> for more information.
    """

    LANGUAGES = {'Elm'}
    REQUIREMENTS = {NpmRequirement('elm', '0.18')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which('elm-format') is None:
            return ('elm-format is missing. Download it from '
                    'https://github.com/avh4/elm-format/blob/master/README.md'
                    '#for-elm-018 and put it into your PATH.')
        else:
            return True

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename, '--yes'



import autoflake

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result


class PyUnusedCodeBear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('autoflake', '0.6.6')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unused Code'}

    def run(self, filename, file,
            remove_all_unused_imports: bool=False,
            remove_unused_variables: bool=True):
        """
        Detects unused code. By default this functionality is limited to:

        - Unneeded pass statements.
        - Unneeded builtin imports.

        :param remove_all_unused_imports:
            True removes all unused imports - might have side effects
        :param remove_unused_variables:
            True removes unused variables - might have side effects
        """

        corrected = autoflake.fix_code(
                       ''.join(file),
                       additional_imports=None,
                       remove_all_unused_imports=remove_all_unused_imports,
                       remove_unused_variables=remove_unused_variables
                       ).splitlines(True)

        for diff in Diff.from_string_arrays(file, corrected).split_diff():
            yield Result(self,
                         'This file contains unused source code.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff})



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='happiness',
        output_format='regex',
        output_regex=r'\s.+:(?P<line>\d+):(?P<column>\d+):(?P<message>.+)')
class HappinessLintBear:
    """
    Checks JavaScript files for semantic and syntax errors using ``happiness``.

    See <https://github.com/JedWatson/happiness/> for more information.
    """
    LANGUAGES = {'JavaScript'}
    REQUIREMENTS = {NpmRequirement('happiness', '7.1.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/80714'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='rstcheck',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.+\s?:(?P<line>\d+):\s'
                     r'(?P<message>\((?P<severity>\w+).+)',
        severity_map={'INFO': RESULT_SEVERITY.INFO,
                      'WARNING': RESULT_SEVERITY.NORMAL,
                      'ERROR': RESULT_SEVERITY.MAJOR,
                      'SEVERE': RESULT_SEVERITY.MAJOR})
class RSTcheckBear:

    """
    Check syntax of ``reStructuredText`` and code blocks
    nested within it.

    Check <https://pypi.python.org/pypi/rstcheck> for more information.
    """

    LANGUAGES = {'reStructuredText'}
    REQUIREMENTS = {PipRequirement('rstcheck', '2.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}
    ASCIINEMA_URL = 'https://asciinema.org/a/8ntlaqubk2qkrn9mm0dh07rlk?speed=2'

    @staticmethod
    def create_arguments(filename, file, config_file,
                         code_block_language_ignore: list=()):
        """
        :param code_block_language_ignore:
            Comma seperated list for languages of which code blocks
            should be ignored. Code block of following languages
            can be detected: ``bash``, ``c``, ``cpp``, ``json``,
            ``python``, ``rst``.
        """
        args = ()
        if code_block_language_ignore:
            args = ('--ignore-language=' +
                    ','.join(code_block_language_ignore),)
        return args + (filename,)



from clang.cindex import Index, CursorKind

from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange
from coalib.bearlib import deprecate_settings
from bears.c_languages.ClangBear import clang_available, ClangBear


class ClangComplexityBear(LocalBear):
    """
    Calculates cyclomatic complexity of each function and displays it to the
    user.
    """

    LANGUAGES = ClangBear.LANGUAGES
    REQUIREMENTS = ClangBear.REQUIREMENTS
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Complexity'}

    check_prerequisites = classmethod(clang_available)
    _decisive_cursor_kinds = {
        CursorKind.IF_STMT, CursorKind.WHILE_STMT, CursorKind.FOR_STMT,
        CursorKind.DEFAULT_STMT, CursorKind.CASE_STMT}

    def function_key_points(self, cursor, top_function_level=False):
        """
        Calculates number of function's decision points and exit points.

        :param top_function_level: Whether cursor is in the top level of
                                   the function.
        """
        decisions, exits = 0, 0

        for child in cursor.get_children():
            if child.kind in self._decisive_cursor_kinds:
                decisions += 1
            elif child.kind == CursorKind.RETURN_STMT:
                exits += 1
                if top_function_level:
                    # There is no point to move forward, so just return.
                    return decisions, exits
            child_decisions, child_exits = self.function_key_points(child)
            decisions += child_decisions
            exits += child_exits

        if top_function_level:
            # Implicit return statement.
            exits += 1

        return decisions, exits

    def complexities(self, cursor, filename):
        """
        Calculates cyclomatic complexities of functions.
        """

        file = cursor.location.file

        if file is not None and file.name != filename:
            # There is nothing to do in another file.
            return

        if cursor.kind == CursorKind.FUNCTION_DECL:
            child = next((child for child in cursor.get_children()
                          if child.kind != CursorKind.PARM_DECL),
                         None)
            if child:
                decisions, exits = self.function_key_points(child, True)
                complexity = max(1, decisions - exits + 2)
                yield cursor, complexity
        else:
            for child in cursor.get_children():
                yield from self.complexities(child, filename)

    @deprecate_settings(cyclomatic_complexity='max_complexity')
    def run(self, filename, file, cyclomatic_complexity: int=8):
        """
        Check for all functions if they are too complicated using the
        cyclomatic complexity metric.

        You can read more about this metric at
        <https://www.wikiwand.com/en/Cyclomatic_complexity>.

        :param cyclomatic_complexity:  Maximum cyclomatic complexity that is
                                considered to be normal. The value of 10 had
                                received substantial corroborating evidence.
                                But the general recommendation: "For each
                                module, either limit cyclomatic complexity to
                                [the agreed-upon limit] or provide a written
                                explanation of why the limit was exceeded."
        """

        root = Index.create().parse(filename).cursor
        for cursor, complexity in self.complexities(root, filename):
            if complexity > cyclomatic_complexity:
                affected_code = (SourceRange.from_clang_range(cursor.extent),)
                yield Result(
                    self,
                    "The function '{function}' should be simplified. Its "
                    'cyclomatic complexity is {complexity} which exceeds '
                    'maximal recommended value '
                    'of {rec_value}.'.format(
                        function=cursor.displayname,
                        complexity=complexity,
                        rec_value=cyclomatic_complexity),
                    affected_code=affected_code,
                    additional_info=(
                        'The cyclomatic complexity is a metric that measures '
                        'how complicated a function is by counting branches '
                        'and exits of each function.\n\n'
                        'Your function seems to be complicated and should be '
                        'refactored so that it can be understood by other '
                        'people easily.\n\nSee '
                        '<http://www.wikiwand.com/en/Cyclomatic_complexity>'
                        ' for more information.'))



from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import typed_list
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from dependency_management.requirements.AnyOneOfRequirements import (
    AnyOneOfRequirements)
from dependency_management.requirements.ComposerRequirement import (
    ComposerRequirement)


@linter(executable='phpmd',
        output_format='regex',
        output_regex=r':(?P<line>\d+)\s*(?P<message>.*)')
class PHPMessDetectorBear:
    """
    The bear takes a given PHP source code base and looks for several
    potential problems within that source. These problems can be things like:

    - Possible bugs
    - Suboptimal code
    - Overcomplicated expressions
    - Unused parameters, methods, properties
    """

    LANGUAGES = {'PHP'}
    REQUIREMENTS = {
        AnyOneOfRequirements(
            [DistributionRequirement(apt_get='phpmd',
                                     dnf='php-phpmd-PHP-PMD',
                                     ),
             ComposerRequirement('phpmd/phpmd'),
             ],
        ),
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Complexity', 'Unused Code', 'Redundancy',
                  'Variable Misuse'}
    SEE_MORE = 'https://phpmd.org/about.html'

    @staticmethod
    def create_arguments(filename, file, config_file,
                         phpmd_rulesets: typed_list(str)):
        """
        :param phpmd_rulesets:
            A list of rulesets to use for analysis.
            Available rulesets: cleancode, codesize, controversial, design,
            naming, unusedcode.
        """
        return filename, 'text', ','.join(phpmd_rulesets)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.settings.Setting import typed_list


@linter(executable='cppcheck',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'(?P<line>\d+):(?P<severity>[a-zA-Z]+):'
                     r'(?P<origin>[a-zA-Z]+):(?P<message>.*)',
        severity_map={'error': RESULT_SEVERITY.MAJOR,
                      'warning': RESULT_SEVERITY.NORMAL,
                      'style': RESULT_SEVERITY.INFO})
class CPPCheckBear:
    """
    Report possible security weaknesses for C/C++.
    For more information, consult <https://github.com/danmar/cppcheck>.
    """

    LANGUAGES = {'C', 'C++'}
    REQUIREMENTS = {DistributionRequirement('cppcheck')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Security', 'Unused Code', 'Unreachable Code', 'Smell'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         enable: typed_list(str)=[]):
        """
        :param enable:
            Choose specific issues to report. Issues that can be
            reported are: all, warning, style, performance,
            portability, information, unusedFunction,
            missingInclude
        """
        args = ('--template={line}:{severity}:{id}:{message}',)

        if enable:
            args += ('--enable=' + ','.join(enable),)

        return args + (filename,)



import eradicate

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result


class PyCommentedCodeBear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('eradicate', '0.1.6')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Commented Code'}

    def run(self, filename, file):
        """
        Detects commented out source code in Python.
        """
        corrected = tuple(eradicate.filter_commented_out_code(''.join(file)))

        for diff in Diff.from_string_arrays(file, corrected).split_diff():
            yield Result(self,
                         'This file contains commented out source code.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff})



import functools
from itertools import combinations

from bears.c_languages.ClangBear import clang_available, ClangBear
from bears.c_languages.codeclone_detection.ClangCountingConditions import (
    condition_dict)
from bears.c_languages.codeclone_detection.ClangCountVectorCreator import (
    ClangCountVectorCreator)
from bears.c_languages.codeclone_detection.CloneDetectionRoutines import (
    compare_functions, get_count_matrices)
from coala_utils.string_processing.StringConverter import StringConverter
from coalib.bears.GlobalBear import GlobalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.collecting.Collectors import collect_dirs
from coalib.results.HiddenResult import HiddenResult
from coalib.settings.Setting import path_list, typed_ordered_dict

# counting_condition_dict is a function object generated by typed_dict. This
# function takes a setting and creates a dictionary out of it while it
# converts all keys to counting condition function objects (via the
# condition_dict) and all values to floats while unset values default to 1.
counting_condition_dict = typed_ordered_dict(
    lambda setting: condition_dict[str(setting).lower()],
    float,
    1)

default_cc_dict = counting_condition_dict(StringConverter(
    """
used: 0,
returned: 1.4,
is_condition: 0,
in_condition: 1.4,
in_second_level_condition: 1.4,
in_third_level_condition: 1.0,
is_assignee: 0,
is_assigner: 0.6,
loop_content: 0,
second_level_loop_content,
third_level_loop_content,
is_param: 2,
is_called: 1.4,
is_call_param: 0.0,
in_sum: 2.0,
in_product: 0,
in_binary_operation,
member_accessed"""))


def get_difference(function_pair,
                   count_matrices,
                   average_calculation,
                   poly_postprocessing,
                   exp_postprocessing):
    """
    Retrieves the difference between two functions using the munkres algorithm.

    :param function_pair:       A tuple containing both indices for the
                                count_matrices dictionary.
    :param count_matrices:      A dictionary holding CMs.
    :param average_calculation: If set to true the difference calculation
                                function will take the average of all variable
                                differences as the difference, else it will
                                normalize the function as a whole and thus
                                weighting in variables dependent on their size.
    :param poly_postprocessing: If set to true, the difference value of big
                                function pairs will be reduced using a
                                polynomial approach.
    :param exp_postprocessing:  If set to true, the difference value of big
                                function pairs will be reduced using an
                                exponential approach.
    :return:                    A tuple containing both function ids and their
                                difference.
    """
    function_1, function_2 = function_pair
    return (function_1,
            function_2,
            compare_functions(count_matrices[function_1],
                              count_matrices[function_2],
                              average_calculation,
                              poly_postprocessing,
                              exp_postprocessing))


class ClangFunctionDifferenceBear(GlobalBear):
    check_prerequisites = classmethod(clang_available)
    LANGUAGES = ClangBear.LANGUAGES
    REQUIREMENTS = ClangBear.REQUIREMENTS | {PipRequirement('munkres3', '1.0')}

    def run(self,
            counting_conditions: counting_condition_dict=default_cc_dict,
            average_calculation: bool=False,
            poly_postprocessing: bool=True,
            exp_postprocessing: bool=False,
            extra_include_paths: path_list=()):
        '''
        Retrieves similarities for code clone detection. Those can be reused in
        another bear to produce results.

        Postprocessing may be done because small functions are less likely to
        be clones at the same difference value than big functions which may
        provide a better refactoring opportunity for the user.

        :param counting_conditions: A comma seperated list of counting
                                    conditions. Possible values are: used,
                                    returned, is_condition, in_condition,
                                    in_second_level_condition,
                                    in_third_level_condition, is_assignee,
                                    is_assigner, loop_content,
                                    second_level_loop_content,
                                    third_level_loop_content, is_param,
                                    in_sum, in_product, in_binary_operation,
                                    member_accessed.
                                    Weightings can be assigned to each
                                    condition due to providing a dict
                                    value, i.e. having used weighted in
                                    half as much as other conditions would
                                    simply be: "used: 0.5, is_assignee".
                                    Weightings default to 1 if unset.
        :param average_calculation: If set to true the difference calculation
                                    function will take the average of all
                                    variable differences as the difference,
                                    else it will normalize the function as a
                                    whole and thus weighting in variables
                                    dependent on their size.
        :param poly_postprocessing: If set to true, the difference value of big
                                    function pairs will be reduced using a
                                    polynomial approach.
        :param extra_include_paths: A list containing additional include paths.
        :param exp_postprocessing:  If set to true, the difference value of big
                                    function pairs will be reduced using an
                                    exponential approach.
        '''
        self.debug('Using the following counting conditions:')
        for key, val in counting_conditions.items():
            self.debug(' *', key.__name__, '(weighting: {})'.format(val))

        self.debug('Creating count matrices...')
        count_matrices = get_count_matrices(
            ClangCountVectorCreator(list(counting_conditions.keys()),
                                    list(counting_conditions.values())),
            list(self.file_dict.keys()),
            lambda prog: self.debug('{:2.4f}%...'.format(prog)),
            self.section['files'].origin,
            collect_dirs(extra_include_paths))

        self.debug('Calculating differences...')

        differences = []
        function_count = len(count_matrices)
        # Thats n over 2, hardcoded to simplify calculation
        combination_length = function_count * (function_count-1) / 2
        partial_get_difference = functools.partial(
            get_difference,
            count_matrices=count_matrices,
            average_calculation=average_calculation,
            poly_postprocessing=poly_postprocessing,
            exp_postprocessing=exp_postprocessing)

        for i, elem in enumerate(
                map(partial_get_difference,
                    [(f1, f2) for f1, f2 in combinations(count_matrices, 2)])):
            if i % 50 == 0:
                self.debug('{:2.4f}%...'.format(100*i/combination_length))
            differences.append(elem)

        yield HiddenResult(self, differences)
        yield HiddenResult(self, count_matrices)



from shutil import which
import sys

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.settings.Setting import typed_list


@linter(executable=sys.executable,
        output_format='regex',
        output_regex=r'(?P<line>\d+):(?P<column>\d+): '
                     r'(?P<severity>Error|Warning|Info): (?P<message>.+)')
class HTMLLintBear:
    """
    Check HTML source code for invalid or misformatted code.

    See also <https://pypi.python.org/pypi/html-linter>.
    """

    _html_lint = which('html_lint.py')

    LANGUAGES = {'HTML', 'Jinja2', 'PHP'}
    REQUIREMENTS = {PipRequirement('html-linter', '0.3.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         htmllint_ignore: typed_list(str)=()):
        """
        :param htmllint_ignore: List of checkers to ignore.
        """
        ignore = ','.join(part.strip() for part in htmllint_ignore)
        return HTMLLintBear._html_lint, '--disable=' + ignore, filename



import itertools

from coalib.bears.GlobalBear import GlobalBear
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class DuplicateFileBear(GlobalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Duplication'}

    def run(self):
        """
        Checks for Duplicate Files
        """
        if not self.file_dict:
            yield Result(self, 'You did not add any file to compare',
                         severity=RESULT_SEVERITY.MAJOR)
        elif len(self.file_dict) == 1:
            yield Result(self, 'You included only one file',
                         severity=RESULT_SEVERITY.MAJOR)
        else:
            unique_tuples = [file_tuple for file_tuple in
                             itertools.combinations(self.file_dict, 2)]

            for file_pair in unique_tuples:
                if (self.file_dict[file_pair[0]] ==
                        self.file_dict[file_pair[1]]):
                    first_file_name = file_pair[0]
                    second_file_name = file_pair[1]
                    message = ('File ' + first_file_name + ' is identical'
                               ' to File ' + second_file_name)
                    yield Result.from_values(origin=self, message=message,
                                             severity=RESULT_SEVERITY.INFO,
                                             file=first_file_name)



import re
import logging

from coalib.bearlib import deprecate_settings
from coalib.bears.LocalBear import LocalBear
from coalib.results.Diff import Diff
from coalib.results.Result import RESULT_SEVERITY, Result
from coalib.results.SourceRange import SourceRange

from bears.general.AnnotationBear import AnnotationBear


def _get_comments(dependency_results):
    if not dependency_results:
        return

    annotation_bear_results = dependency_results.get('AnnotationBear')
    if (not annotation_bear_results or
            not isinstance(annotation_bear_results, list)):
        return

    for result in annotation_bear_results:
        if isinstance(result.contents, str):
            logging.error(result.contents)
        else:
            yield from result.contents.get('comments', [])


def generate_diff(comments, file, filename,
                  line, line_number, pos):
    todo_source_range = SourceRange.from_values(filename, line_number,
                                                pos + 1)
    affected_comment_sourcerange = [
        c for c in comments if todo_source_range in c]

    affected_len = len(affected_comment_sourcerange)

    if affected_len == 0:
        return {}
    assert affected_len == 1, 'More than 1 affected comment source ranges'

    comment_sourcerange = affected_comment_sourcerange[0]

    comment_start = comment_sourcerange.start.column
    comment_end = comment_sourcerange.end.column
    in_multi_line_comment = (
        comment_sourcerange.start.line != comment_sourcerange.end.line
    )

    line_before_todo_comment = line[:comment_start - 1].rstrip()
    line_behind_todo_comment = line[comment_end:].rstrip()

    line_replacement = line_before_todo_comment + line_behind_todo_comment

    diff = Diff(file)
    if line_replacement and not in_multi_line_comment:
        diff.change_line(
            line_number,
            line,
            line_replacement + '\n')
    elif line_replacement and in_multi_line_comment:
        text_replacement = line[pos:]
        diff.change_line(
            line_number,
            line,
            line.replace(text_replacement, '').rstrip() + '\n')
    else:
        diff.delete_line(line_number)

    return {filename: diff}


class KeywordBear(LocalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Documentation'}
    BEAR_DEPS = {AnnotationBear}

    @deprecate_settings(keywords='ci_keywords')
    def run(self,
            filename,
            file,
            keywords: list=['todo', 'fixme'],
            regex_keyword: str='',
            dependency_results: dict=None):
        '''
        Checks the code files for given keywords.

        :param keywords:
            A list of keywords to search for (case insensitive).
            Default are TODO and FIXME.
        :param regex_keyword:
            A regular expression to search for matching keywords in a file.
        '''
        comments = list(_get_comments(dependency_results))

        if keywords:
            simple_keywords_regex = re.compile(
                '(' + '|'.join(re.escape(key) for key in keywords) + ')',
                re.IGNORECASE)

            message = "The line contains the keyword '{}'."
            yield from self.check_keywords(filename, file, comments,
                                           simple_keywords_regex, message)

        if regex_keyword is not '':
            regex = re.compile(regex_keyword)
            message = ("The line contains the keyword '{}' which "
                       'resulted in a match with given regex.')
            yield from self.check_keywords(filename, file, comments, regex,
                                           message)

    def check_keywords(self,
                       filename,
                       file,
                       comments,
                       regex,
                       message):
        '''
        Checks for the presence of keywords according to regex in a given file.

        :param regex:
           A regular expression which is used to search matching
           keywords in a file.
        :param message:
           A message to be displayed to the user when a keyword in a given file
           results in a match. It may have an unnamed placeholder for the
           keyword.
        '''

        for line_number, line in enumerate(file, start=1):
            for keyword in regex.finditer(line):
                diffs = generate_diff(
                    comments,
                    file,
                    filename,
                    line,
                    line_number,
                    keyword.start())
                yield Result.from_values(
                    origin=self,
                    message=message.format(keyword.group()),
                    file=filename,
                    line=line_number,
                    column=keyword.start() + 1,
                    end_line=line_number,
                    end_column=keyword.end() + 1,
                    severity=RESULT_SEVERITY.INFO,
                    diffs=diffs)



from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import path
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


_online_styles = {
    'android-check-easy': 'https://raw.githubusercontent.com/noveogroup/' +
    'android-check/master/android-check-plugin/src/main/resources/' +
    'checkstyle/checkstyle-easy.xml',
    'android-check-hard': 'https://raw.githubusercontent.com/noveogroup/' +
    'android-check/master/android-check-plugin/src/main/resources/' +
    'checkstyle/checkstyle-hard.xml',
}

# To be deprecated
known_checkstyles = dict(_online_styles, **{'google': None, 'sun': None})


def check_invalid_configuration(checkstyle_configs, use_spaces, indent_size):
    if (checkstyle_configs is 'google' and
            (not use_spaces or indent_size != 2)):
        raise ValueError('Google checkstyle config does not support '
                         'use_spaces=False or indent_size != 2')


def known_checkstyle_or_path(setting):
    if str(setting) in known_checkstyles.keys():
        return str(setting)
    else:
        return path(setting)


@linter(executable='java',
        output_format='regex',
        output_regex=r'\[(?P<severity>ERROR|WARN|INFO)\].*?'
                     r'(?P<line>\d+):?(?P<column>\d+)?. '
                     r'(?P<message>.*?) *\[(?P<origin>[a-zA-Z]+?)\]')
class CheckstyleBear:
    """
    Check Java code for possible style, semantic and design issues.

    For more information, consult
    <http://checkstyle.sourceforge.net/checks.html>.
    """

    LANGUAGES = {'Java'}
    REQUIREMENTS = {DistributionRequirement(apt_get='default-jre')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Smell'}

    def setup_dependencies(self):
        type(self).checkstyle_jar_file = self.download_cached_file(
            'http://sourceforge.net/projects/checkstyle/files/checkstyle/6.19'
            '/checkstyle-6.19-all.jar',
            'checkstyle-6.19.jar')

    def create_arguments(
            self, filename, file, config_file,
            checkstyle_configs: known_checkstyle_or_path='google',
            use_spaces: bool=True, indent_size: int=2):
        """
        :param checkstyle_configs:
            A file containing configs to use in ``checkstyle``. It can also
            have the special values:

            - google - Google's Java style. More info at
              <http://checkstyle.sourceforge.net/style_configs.html>.
            - sun - Sun's Java style. These are the same
              as the default Eclipse checks. More info at
              <http://checkstyle.sourceforge.net/style_configs.html>.
            - android-check-easy - The easy Android configs provided by the
              android-check eclipse plugin. More info at
              <https://github.com/noveogroup/android-check>.
            - android-check-hard - The hard Android confis provided by the
              android-check eclipse plugin. More info at
              <https://github.com/noveogroup/android-check>.
            - geosoft - The Java style followed by GeoSoft. More info at
              <http://geosoft.no/development/javastyle.html>
        """
        check_invalid_configuration(
            checkstyle_configs, use_spaces, indent_size)

        if checkstyle_configs in ('google', 'sun'):
            # Checkstyle included these two rulesets since version 6.2
            # Locate the file as an absolute resource into the checkstyle jar
            checkstyle_configs = '/%s_checks.xml' % checkstyle_configs
        elif checkstyle_configs in _online_styles:
            checkstyle_configs = self.download_cached_file(
                _online_styles[checkstyle_configs],
                checkstyle_configs + '.xml')

        return ('-jar', self.checkstyle_jar_file, '-c',
                checkstyle_configs, filename)



from shutil import which
from xml.etree import ElementTree

from coalib.bears.GlobalBear import GlobalBear
from coalib.misc.Shell import run_shell_command
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange


class CPDBear(GlobalBear):

    language_dict = {'C#': 'cs',
                     'C++': 'cpp',
                     'JavaScript': 'ecmascript',
                     'Fortran': 'fortran',
                     'Go': 'go',
                     'Java': 'java',
                     'JSP': 'jsp',
                     'Matlab': 'matlab',
                     'Octave': 'matlab',
                     'Objective-C': 'objectivec',
                     'PHP': 'php',
                     'PL/SQL': 'plsql',
                     'Python': 'python',
                     'Python 2': 'python',
                     'Python 3': 'python',
                     'Ruby': 'ruby',
                     'Scala': 'scala',
                     'Swift': 'swift'}

    lowered_lang_dict = {key.lower(): value
                         for key, value in language_dict.items()}

    LANGUAGES = set(language_dict.keys())
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Duplication'}

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which('bash') is None:
            return 'bash is not installed.'
        if which('pmd') is None and which('run.sh') is None:
            return ('PMD is missing. Make sure to install it from '
                    '<https://pmd.github.io/>.')
        else:
            return True

    def run(self, language: str, minimum_tokens: int=20,
            ignore_annotations: bool=False, ignore_identifiers: bool=True,
            ignore_literals: bool=False, ignore_usings: bool=False,
            skip_duplicate_files: bool=True):
        """
        Checks for similar code that looks as it could be replaced to reduce
        redundancy.

        For more details see:
        <https://pmd.github.io/pmd-5.4.1/usage/cpd-usage.html>

        :param language:
            One of the supported languages of this bear.
        :param minimum_tokens:
            The minimum token length which should be reported as a duplicate.
        :param ignore_annotations:
            Ignore language annotations when comparing text.
        :param ignore_identifiers:
            Ignore constant and variable names when comparing text.
        :param ignore_literals:
            Ignore number values and string contents when comparing text.
        :param ignore_usings:
            Ignore ``using`` directives in C#.
        :param skip_duplicate_files:
            Ignore multiple copies of files of the same name and length in
            comparison.
        """
        language = language.lower()

        if language not in self.lowered_lang_dict:  # pragma: no cover
            self.err('This bear does not support files with the extension '
                     "'{}'.".format(language))
            return

        options = {
            '--ignore-annotations': ignore_annotations,
            '--ignore-identifiers': ignore_identifiers,
            '--ignore-literals': ignore_literals,
            '--ignore-usings': ignore_usings,
            '--skip-duplicate-files': skip_duplicate_files}

        files = ','.join(self.file_dict.keys())
        executable = which('pmd') or which('run.sh')
        arguments = ('bash', executable, 'cpd', '--skip-lexical-errors',
                     '--minimum-tokens', str(minimum_tokens),
                     '--language', self.lowered_lang_dict[language],
                     '--files', files,
                     '--format', 'xml')

        arguments += tuple(option
                           for option, enable in options.items()
                           if enable is True)

        stdout_output, _ = run_shell_command(arguments)

        if stdout_output:
            root = ElementTree.fromstring(stdout_output)

            for duplication in root.findall('duplication'):
                length = int(duplication.attrib['lines'])
                affected_code = list()

                for xml_file in duplication.findall('file'):
                    filename = xml_file.attrib['path']
                    start_line = int(xml_file.attrib['line'])
                    end_line = min(start_line + length - 1,
                                   len(self.file_dict[filename]))

                    affected_code.append(
                        SourceRange.from_values(filename,
                                                start_line=start_line,
                                                end_line=end_line))

                yield Result(
                    self, 'Duplicate code found.', affected_code,
                    additional_info=(
                        'Duplicate code is an indicator '
                        'that you have more code than you need. Consider'
                        ' refactor your code to remove one of the'
                        ' occurrences. For more information go here:'
                        'http://tinyurl.com/coala-clone'))



from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from dependency_management.requirements.PipRequirement import PipRequirement
import yaml


@linter(executable='yamllint',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): '
                     r'\[(?P<severity>error|warning)\] (?P<message>.+)',
        severity_map={'error': RESULT_SEVERITY.MAJOR,
                      'warning': RESULT_SEVERITY.NORMAL})
class YAMLLintBear:
    """
    Check yaml code for errors and possible problems.

    You can read more about capabilities at
    <http://yamllint.readthedocs.org/en/latest/rules.html>.
    """

    LANGUAGES = {'YAML'}
    REQUIREMENTS = {PipRequirement('yamllint', '1.6.1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def generate_config(filename, file,
                        document_start: bool=None,
                        max_line_length: int=80):
        """
        :param document_start:
            Use this rule to require or forbid the use of document start
            marker (---).
        :param max_line_length:
            Maximum number of characters for a line, the newline character
            being excluded.
        """
        yamllint_configs = {
            'extends': 'default',
            'rules': {
                'document-start': 'disable' if document_start is None
                                  else {'present': document_start},
                'line-length': {
                    'max': max_line_length,
                    },
            },
        }

        return yaml.dump(yamllint_configs)

    @staticmethod
    def create_arguments(filename, file, config_file, yamllint_config: str=''):
        """
        :param yamllint_config: Path to a custom configuration file.
        """
        args = ('-f', 'parsable', filename)
        if yamllint_config:
            args += ('--config-file=' + yamllint_config,)
        else:
            args += ('--config-file=' + config_file,)
        return args



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='pyflakes',
        use_stderr=True,
        output_format='regex',
        output_regex=r'.*:(?P<line>\d+):'
                     r'[(?P<column>\d+):|?]*(?P<severity>)\s(?P<message>.*)\n',
        severity_map={
            '': RESULT_SEVERITY.INFO
        })
class PyFlakesBear:
    """
    Checks Python files for errors using ``pyflakes``.

    See https://github.com/PyCQA/pyflakes for more info.
    """
    LANGUAGES = {'Python', 'Python 3'}
    REQUIREMENTS = {PipRequirement('pyflakes', '1.4.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/92503'
    CAN_DETECT = {'Syntax', 'Unused Code', 'Undefined Element'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='dartanalyzer',
        output_format='regex',
        output_regex=r'\[(?P<severity>error|warning)\] (?P<message>.+)\('
                     r'.+, line (?P<line>\d+), col (?P<column>\d+)\)')
class DartLintBear:
    """
    Checks the code with ``dart-linter``.

    This bear expects dart commands to be on your ``PATH``. Please ensure
    /path/to/dart-sdk/bin is in your ``PATH``.
    """
    LANGUAGES = {'Dart'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(
            filename, file, config_file,
            use_spaces: bool=True, indent_size: int=2):

        # use_spaces must be True and indent_size must be 2 because
        # dartanalyzer only supports these settings
        # see https://github.com/dart-lang/dart_style/issues/261
        if (indent_size != 2 or not use_spaces):
            raise ValueError(
                'DartLintBear only supports `use_spaces=True` '
                'and `indent_size=2`'
            )
        return filename,



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='dockerfile_lint')
class DockerfileLintBear:
    """
    Check file syntax as well as arbitrary semantic and best practice
    in Dockerfiles. it also checks LABEL rules against docker images.

    Uses ``dockerfile_lint`` to provide the analysis.
    See <https://github.com/projectatomic/dockerfile_lint#dockerfile-lint> for
    more information .
    """
    LANGUAGES = {'Dockerfile'}
    REQUIREMENTS = {NpmRequirement('dockerfile_lint', '0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Smell'}

    severity_map = {
        'error': RESULT_SEVERITY.MAJOR,
        'warn': RESULT_SEVERITY.NORMAL,
        'info': RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--json', '-f', filename

    def process_output(self, output, filename, file):
        output = json.loads(output)

        for severity in output:
            if severity == 'summary':
                continue
            for issue in output[severity]['data']:
                yield Result.from_values(
                    origin=self,
                    message=issue['message'],
                    file=filename,
                    severity=self.severity_map[issue['level']],
                    line=issue.get('line'))



"""
coala-bears is a Python package containing all
the bears that are officially supported by coala.
"""


from distutils.version import StrictVersion
import logging
import os
import sys


VERSION_FILE = os.path.join(os.path.dirname(__file__), 'VERSION')
with open(VERSION_FILE, 'r') as ver:
    VERSION = ver.readline().strip()


__version__ = VERSION


def assert_supported_version():  # pragma: no cover
    if sys.version_info < (3, 4):
        print('coala supports only python 3.4 or later.')
        exit(4)


def check_coala_version():
    """
    Check the installed coala and coala-bears version.
    """
    try:
        import coalib
        # check only MAJOR.MINOR version
        coalib_version = '.'.join(coalib.__version__.split('.')[:2])
        bears_version = '.'.join(__version__.split('.')[:2])
        if StrictVersion(coalib_version) > StrictVersion(bears_version):
            logging.warning('Version mismatch between coala ({}) and '
                            'coala-bears ({}). This may or may not cause '
                            'errors. If you encounter a problem, try to '
                            'update coala and coala-bears to latest version '
                            "with 'pip3 install -U coala coala-bears'."
                            .format(coalib.__version__, __version__))
    except ImportError:  # pragma: no cover
        logging.error('Module coalib is not found. Cannot do version '
                      'checking.')


check_coala_version()



from coala_utils.string_processing.Core import escape
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='julia',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+) (?P<severity>.)\d+ (?P<message>.*)',
        severity_map={'E': RESULT_SEVERITY.MAJOR,
                      'W': RESULT_SEVERITY.NORMAL,
                      'I': RESULT_SEVERITY.INFO},
        prerequisite_check_command=('julia', '-e', 'import Lint.lintfile'),
        prerequisite_check_fail_message='Lint package not installed. Run '
                                        '`Pkg.add("Lint")` from Julia to '
                                        'install Lint.')
class JuliaLintBear:
    """
    Provide analysis related to common bugs and potential issues in Julia like
    dead code, undefined variable usage, duplicate keys in dicts, incorrect
    ADT usage, wrongfully using ellipsis, and much more.

    See <https://lintjl.readthedocs.org/en/stable/> for more information
    on the analysis provided.
    """
    LANGUAGES = {'Julia'}
    REQUIREMENTS = {DistributionRequirement(apt_get='julia')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unused Code', 'Syntax', 'Redundancy', 'Duplication',
                  'Unreachable Code', 'Security', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        lintcode = ('import Lint.lintfile; display(lintfile("' +
                    escape(filename, '"\\') + '"))')
        return '-e', lintcode
